# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from earthkit.data.utils import ensure_dict
from earthkit.data.utils import ensure_iterable
from earthkit.data.utils.dates import datetime_from_grib

LOG = logging.getLogger(__name__)


class CompoundKey:
    name = None
    keys = []

    def remapping(self):
        return {self.name: "".join([str(k) for k in self.keys])}

    @staticmethod
    def make(key):
        ck = COMPOUND_KEYS.get(key, None)
        return ck() if ck is not None else None


class ParamLevelKey(CompoundKey):
    name = "param_level"
    keys = ["param", "level", "levelist"]


COMPOUND_KEYS = {v.name: v for v in [ParamLevelKey]}


ENS_KEYS = ["number", "perturbationNumber", "realization"]
LEVEL_KEYS = ["level", "levelist", "topLevel", "bottomLevel", "levels"]
LEVEL_TYPE_KEYS = ["typeOfLevel", "levtype"]
DATE_KEYS = ["date", "andate", "validityDate", "dataDate", "hdate", "referenceDate", "indexingDate"]
TIME_KEYS = ["time", "antime", "validityTime", "dataTime", "referenceTime", "indexingTime"]
STEP_KEYS = ["step_timedelta", "step", "endStep", "stepRange", "forecastMonth", "fcmonth"]
MONTH_KEYS = ["forecastMonth", "fcmonth"]
VALID_DATETIME_KEYS = ["valid_time", "valid_datetime"]
BASE_DATETIME_KEYS = [
    "forecast_reference_time",
    "base_time",
    "base_datetime",
    "reference_time",
    "reference_datetime",
    "indexing_time",
    "indexing_datetime",
]
DATETIME_KEYS = BASE_DATETIME_KEYS + VALID_DATETIME_KEYS

KEYS = (
    ENS_KEYS
    + LEVEL_KEYS
    + LEVEL_TYPE_KEYS
    + DATE_KEYS
    + TIME_KEYS
    + STEP_KEYS
    + VALID_DATETIME_KEYS
    + BASE_DATETIME_KEYS
)


def get_keys(keys, drop=None):
    r = list(keys)
    if drop:
        drop = ensure_iterable(drop)
        r = [k for k in r if k not in drop]
    return r


def find_alias(key, drop=None):
    keys = KEYS
    r = []
    for k in keys:
        if key in k:
            r.extend(k)

    if drop:
        drop = ensure_iterable(drop)
        r = [k for k in r if k not in drop]
    return r


def make_dim(owner, *args, name=None, key=None, **kwargs):
    predef_key = key or name

    if predef_key in PREDEFINED_DIMS:
        return PREDEFINED_DIMS[predef_key](owner, *args, name=name, key=key, **kwargs)

    ck = CompoundKey.make(name)
    if ck is not None:
        d = CompoundKeyDim(owner, ck)
    else:
        # print("args", args, "kwargs", kwargs, "name", name, "key", key)
        d = OtherDim(owner, *args, name=name, key=key, **kwargs)
    return d


class Dim:
    """Represent a dimension.

    Parameters
    ----------
    name: str
        Name of the dimension.
    key: str
        Metadata key to be used to get the values of the dimension.
    alias: list
        List of metadata keys that has the same meaning as the ``key``.
    drop: list
        List of metadata keys used to identify the dimensions that cannot be used when the
        current dimension is ``active``. When this dimension is active and the dimension
        consistency is checked all the other dimensions containing any keys from ``drop``
        as ``name``, ``key`` or ``alias`` should be deactivated.
    active: bool
        Status. Only active dimensions are used in the dataset construction.
    """

    name = None
    key = None
    label = None
    alias = None
    drop = None
    enforce_unique = False

    def __init__(self, owner, name=None, key=None, alias=None, drop=None, active=True):
        self.owner = owner
        self.profile = owner.profile
        self.active = active

        if name is not None:
            self.name = name

        if key is not None:
            self.key = key

        if not self.key and self.name:
            self.key = self.name

        if not self.name and self.key:
            self.name = self.key

        if not self.name:
            raise ValueError("name must be specified")

        if not self.key:
            raise ValueError("key must be specified")

        if alias is not None:
            self.alias = alias

        if drop is not None:
            self.drop = drop

        def _drop(v):
            return [k for k in v if k not in [self.name, self.key]]

        self.alias = _drop(ensure_iterable(self.alias))
        self.drop = _drop(ensure_iterable(self.drop))

        assert self.name not in self.alias
        assert self.name not in self.drop
        assert self.key not in self.alias
        assert self.key not in self.drop

        self.coords = {}

    def __contains__(self, key):
        return key in [self.name, self.key] or key in self.alias

    def check(self):
        # print(f"CHECK {self.name} {self.key} {self.active}")
        if self.active:
            self.active = self.condition()
            if self.active:
                self.deactivate_drop_list()

    def condition(self):
        return True

    def update(self, ds):
        # if self.key in ds.indices():
        #     print(f"-> {self.name} key={self.key} active={self.active} ds={ds.index(self.key)}")

        if not self.active:
            return

        # sanity check
        if self.profile.variable.key in self:
            raise ValueError(
                (f"Variable key {self.profile.variable.key} cannot be in " f"dimension={self.name}")
            )

        # print(f"key={self.key} index={ds.index(self.key)}")

        vals = ds.index(self.key)

        if len(vals) == 0:
            self.active = False

        # assert self.name in self.profile.dim_keys, f"self.name={self.name}"
        # if self.key not in self.owner.ensure_dims:
        #     num = len(ds.index(self.key))
        #     if num == 0 or (self.key not in self.owner.ensure_dims and num == 1 and squeeze):
        #         self.active = False

        # self.deactivate_drop_list()

    def deactivate_drop_list(self):
        self.owner.deactivate([self.name, self.key] + self.drop, ignore_dim=self)

    def as_coord(self, key, values, component, source):
        if key not in self.coords:
            from .coord import Coord

            self.coords[key] = Coord.make(key, values, ds=source, component=component)
        return key, self.coords[key]

    def remapping_keys(self):
        return []

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, key={self.key})"


class NumberDim(Dim):
    alias = get_keys(ENS_KEYS)


class DateDim(Dim):
    name = "date"
    drop = get_keys(DATE_KEYS + DATETIME_KEYS, drop="date")


class TimeDim(Dim):
    name = "time"
    drop = get_keys(TIME_KEYS + DATETIME_KEYS, drop="time")


# class StepDim(Dim):
#     name = "step"
#     drop = get_keys(STEP_KEYS + VALID_DATETIME_KEYS, drop="step")


class StepDim(Dim):
    name = "step"
    drop = get_keys(STEP_KEYS + VALID_DATETIME_KEYS, drop=["step_timedelta"])


class ValidTimeDim(Dim):
    name = "valid_time"
    drop = get_keys(DATE_KEYS + TIME_KEYS + DATETIME_KEYS, drop="valid_time")


class ForecastRefTimeDim(Dim):
    name = "forecast_reference_time"
    drop = get_keys(DATE_KEYS + TIME_KEYS + DATETIME_KEYS, drop="forecast_reference_time")
    alias = ["base_datetime"]


class IndexingTimeDim(Dim):
    name = "indexing_time"
    drop = get_keys(DATE_KEYS + TIME_KEYS + DATETIME_KEYS, drop="indexing_time")


class ReferenceTimeDim(Dim):
    name = "reference_time"
    drop = get_keys(DATE_KEYS + TIME_KEYS + DATETIME_KEYS, drop="reference_time")


class CustomForecastRefDim(Dim):
    @staticmethod
    def _datetime(val):
        if not val:
            return None
        else:
            try:
                date, time = val.split("_")
                return datetime_from_grib(int(date), int(time)).isoformat()
            except Exception:
                return val

    def __init__(self, owner, keys, *args, active=True, **kwargs):
        if isinstance(keys, str):
            self.key = keys
        elif isinstance(keys, list) and len(keys) == 2:
            date = keys[0]
            time = keys[1]
            self.key = self._name(date, time)
            self.drop = [date, time]
            if active:
                owner.register_remapping(
                    {self.key: "{" + date + "}_{" + time + "}"},
                    patch={self.key: CustomForecastRefDim._datetime},
                )
        else:
            raise ValueError(f"Invalid keys={keys}")
        super().__init__(owner, *args, active=active, **kwargs)

    def _name(self, date, time):
        if date.endswith("Date") and time.endswith("Time") and date[:-4] == time[:-4]:
            return date[:-4] + "_time"
        return f"{date}_{time}"


class LevelDim(Dim):
    alias = get_keys(LEVEL_KEYS)


class LevelTypeDim(Dim):
    alias = get_keys(LEVEL_TYPE_KEYS)
    enforce_unique = True


class LevelPerTypeDim(Dim):
    name = "_level_per_type"
    drop = get_keys(LEVEL_KEYS + LEVEL_TYPE_KEYS, drop=name)

    def __init__(self, owner, level_key, level_type_key, *args, **kwargs):
        self.key = level_key
        self.level_key = level_key
        self.level_type_key = level_type_key
        super().__init__(owner, *args, **kwargs)

    def as_coord(self, key, values, component, source):
        lev_type = source[0].metadata(self.level_type_key)
        if not lev_type:
            raise ValueError(f"{self.level_type_key} not found in metadata")

        if lev_type not in self.coords:
            from .coord import Coord

            coord = Coord.make(lev_type, list(values), ds=source)
            self.coords[lev_type] = coord
        return lev_type, self.coords[lev_type]


class LevelAndTypeDim(Dim):
    name = "level_and_type"
    drop = get_keys(LEVEL_KEYS + LEVEL_TYPE_KEYS, drop=name)

    def __init__(self, owner, level_key, level_type_key, active=True, *args, **kwargs):
        self.level_key = level_key
        self.level_type_key = level_type_key
        if active:
            owner.register_remapping(
                {self.name: "{" + self.level_key + "}{" + self.level_type_key + "}"},
            )
        super().__init__(owner, *args, active=active, **kwargs)

    def remapping_keys(self):
        return [self.level_key, self.level_type_key]


class RemappingDim(Dim):
    def __init__(self, owner, name, keys, **kwargs):
        self.name = name
        self.keys = [k for k in keys if k]
        self.drop = self.build_drop(keys)
        super().__init__(owner, **kwargs)

    def build_drop(self, keys):
        r = list(keys)
        for k in keys:
            r.extend(find_alias(k))
        return r

    def remapping_keys(self):
        return self.keys


class CompoundKeyDim(RemappingDim):
    def __init__(self, owner, ck):
        self.ck = ck
        super().__init__(owner, ck.name, ck.keys)


class OtherDim(Dim):
    pass


class DimRole:
    NAMES = ("number", "date", "time", "step", "level", "level_type", "forecast_reference_time", "valid_time")

    def __init__(self, d, name_as_key=True):
        self.d = d
        self.name_as_key = name_as_key

        if "ens" in d:
            import warnings

            warnings.warn("'ens' key in dim_roles is deprecated. Use 'number' instead", DeprecationWarning)
            self.d["number"] = self.d.pop("ens")

        for k in d:
            if k not in self.NAMES:
                raise ValueError(f"Invalid dim role name={k}. Must be one of {self.NAMES}")

    def role(self, name, default=None, raise_error=True):
        if name in self.d:
            return self.d[name], name if self.name_as_key else self.d[name]
        if default is not None:
            return default
        if raise_error:
            raise ValueError(f"Dim role {name} not found in {self.d}")
        else:
            return default, default


class DimMode:
    default = {}  # maps key to name

    def build(self, profile, owner, active=True, dims=None):
        if not dims:
            dims = self.default
        return {name: make_dim(owner, name=name, key=key, active=active) for name, key in dims.items()}


class ForecastTimeDimMode(DimMode):
    name = "forecast"
    DATES = ["date", "dataDate"]
    TIMES = ["time", "dataTime"]

    def build(self, profile, owner, active=True):
        ref_time_key, ref_time_name = owner.dim_roles.role("forecast_reference_time", raise_error=False)

        if ref_time_key == "forecast_reference_time":
            ref_time_dim = ForecastRefTimeDim(owner, name=ref_time_name, active=active)
        elif ref_time_key:
            ref_time_dim = make_dim(owner, name=ref_time_name, key=ref_time_key, active=active)
        else:
            date, _ = owner.dim_roles.role("date")
            time, _ = owner.dim_roles.role("time")
            built_in = date in self.DATES and time in self.TIMES
            if built_in:
                ref_time_dim = ForecastRefTimeDim(owner, name=ref_time_name, active=active)
            else:
                ref_time_dim = CustomForecastRefDim(owner, [date, time], name=ref_time_name, active=active)

        step_key, step_name = owner.dim_roles.role("step")
        step_dim = make_dim(owner, name=step_name, key=step_key, active=active)

        self.register_ref_time_key(ref_time_dim.key)
        self.register_step_key(step_dim.key)

        return {d.name: d for d in [ref_time_dim, step_dim]}

    def register_ref_time_key(self, name):
        global BASE_DATETIME_KEYS
        global VALID_DATETIME_KEYS
        if name not in BASE_DATETIME_KEYS:
            BASE_DATETIME_KEYS.append(name)

    def register_step_key(self, name):
        global STEP_KEYS
        if name not in STEP_KEYS:
            STEP_KEYS.append(name)


class ValidTimeDimMode(DimMode):
    name = "valid_time"
    default = {"valid_time": "valid_time"}


class RawTimeDimMode(DimMode):
    name = "raw"

    def build(self, profile, owner, active=True):
        dims = {}
        for k in ["date", "time", "step"]:
            key, name = owner.dim_roles.role(k)
            dims[name] = key
        return super().build(profile, owner, active=active, dims=dims)


class LevelDimMode(DimMode):
    name = "level"

    def build(self, profile, owner, **kwargs):
        # level
        key, name = owner.dim_roles.role("level")
        level_dim = LevelDim(owner, name=name, key=key, **kwargs)

        # level_type
        key, name = owner.dim_roles.role("level_type")
        level_type_dim = LevelTypeDim(owner, name=name, key=key, **kwargs)

        return {level_dim.key: level_dim, level_type_dim.key: level_type_dim}


class LevelAndTypeDimMode(DimMode):
    name = "level_and_type"
    dim = LevelAndTypeDim

    def build(self, profile, owner, **kwargs):

        level_key, _ = owner.dim_roles.role("level")
        level_type_key, _ = owner.dim_roles.role("level_type")
        return {self.name: self.dim(owner, level_key, level_type_key, **kwargs)}


class LevelPerTypeDimMode(LevelAndTypeDimMode):
    name = "level_per_type"
    dim = LevelPerTypeDim


TIME_DIM_MODES = {v.name: v for v in [ForecastTimeDimMode, ValidTimeDimMode, RawTimeDimMode]}
LEVEL_DIM_MODES = {v.name: v for v in [LevelDimMode, LevelPerTypeDimMode, LevelAndTypeDimMode]}


class DimBuilder:
    used = {}
    ignored = {}

    def dims(self):
        return self.used, self.ignored


class NumberDimBuilder(DimBuilder):
    name = "number"

    def __init__(self, profile, owner):
        key, name = owner.dim_roles.role("number")
        self.used = {self.name: NumberDim(owner, name=name, key=key)}


class TimeDimBuilder(DimBuilder):
    name = "time"

    def __init__(self, profile, owner):
        mode = TIME_DIM_MODES.get(owner.time_dim_mode, None)
        if mode is None:
            raise ValueError(f"Unknown time_dim_mode={owner.time_dim_mode}")

        mode = mode()
        self.used = mode.build(profile, owner)
        self.ignored = {
            k: v().build(profile, owner, active=False) for k, v in TIME_DIM_MODES.items() if v != mode
        }


class LevelDimBuilder(DimBuilder):
    name = "level"

    def __init__(self, profile, owner):
        mode = LEVEL_DIM_MODES.get(owner.level_dim_mode, None)
        if mode is None:
            raise ValueError(f"Unknown level_dim_mode={owner.level_dim_mode}")

        mode = mode()
        self.used = mode.build(profile, owner)
        self.ignored = {
            k: v().build(profile, owner, active=False) for k, v in LEVEL_DIM_MODES.items() if v != mode
        }


DIM_BUILDERS = {v.name: v for v in [NumberDimBuilder, TimeDimBuilder, LevelDimBuilder]}


def ensure_dim_map(d):
    if isinstance(d, dict):
        return d
    d = ensure_iterable(d)
    r = {}
    for k in d:
        if isinstance(k, str):
            r[k] = k
        elif isinstance(k, tuple) and len(k) == 2:
            r[k[0]] = k[1]
        elif isinstance(k, dict):
            for kk, vv in k.items():
                r[kk] = vv
    return r


class DimHandler:
    def __init__(
        self,
        profile,
        extra_dims,
        drop_dims,
        ensure_dims,
        fixed_dims,
        split_dims,
        rename_dims,
        dim_roles,
        dim_name_from_role_name,
        dims_as_attrs,
        time_dim_mode,
        level_dim_mode,
        squeeze,
    ):

        self.profile = profile

        self.dim_roles = DimRole(dim_roles, name_as_key=dim_name_from_role_name)
        # self.dim_name_from_role_name = dim_name_from_role_name
        self.extra_dims = ensure_dim_map(extra_dims)
        self.drop_dims = ensure_iterable(drop_dims)
        self.ensure_dims = ensure_iterable(ensure_dims)
        self.fixed_dims = ensure_dim_map(fixed_dims)
        self.split_dims = ensure_iterable(split_dims)
        self.rename_dims_map = ensure_dict(rename_dims)
        self.dims_as_attrs = list(ensure_iterable(dims_as_attrs))
        self.time_dim_mode = time_dim_mode
        self.level_dim_mode = level_dim_mode
        self.squeeze = squeeze

        # if "ens" in self.dim_roles:
        #     Warning.deprecated("'ens' key in dim_roles is deprecated. Use 'number' instead")
        #     self.dim_roles["number"] = self.dim_roles.pop("ens")

        # if self.dim_name_from_role_name:
        #     d = {v: k for k, v in self.dim_roles.items()}
        #     for k in list(self.rename_dims_map.keys()):
        #         if k in self.dim_roles:
        #             d[self.dim_roles[k]] = self.rename_dims_map.pop(k)

        #     d.update(self.rename_dims_map)
        #     self.rename_dims_map = d

        self.var_key_dim = None

        if self.fixed_dims:
            self.dims = self._init_fixed_dims()
        else:
            dims = self._init_dims()

            assert self.var_key_dim

            # prepend the variable key as dimension to ensure it will be used
            # in the checks
            assert "__var_key_dim__" not in dims
            self.dims = {"__var_key_dim__": self.var_key_dim, **dims}

            # print(f"INIT dims={self.dims}")
            # check dims consistency. The ones that can be used
            # marked as active
            for k, d in self.dims.items():
                d.check()

            var_keys = ["__var_key_dim__"]
            if self.profile.variable.key:
                var_keys.append(self.profile.variable.key)

            # print(f" -> dims={self.dims}")
            # check for any dimensions related to variable keys. These have to
            # be removed from the list of active dims.
            var_dims = self.deactivate(var_keys, others=True, collect=True)
            if var_dims:
                if self.profile.variable.key in var_dims:
                    var_dims.remove(self.profile.variable.key)
                if var_dims:
                    raise ValueError(self.var_dim_found_error_message(var_dims))

            # only the active dims are used
            dims = {k: v for k, v in self.dims.items() if v.active and k not in self.core_dim_order}
            for k in self.core_dim_order:
                if k not in dims and k in self.dims and self.dims[k].active:
                    dims[k] = self.dims[k]

            self.dims = dims

        # LOG.debug(f"self.dims={self.dims}")

        # for d in self.dims.values():
        #     if d.name != d.key:
        #         if d.name in self.rename_dims_map:
        #             self.rename_dims_map[d.key] = d.name
        #         else d
        #         if d.key not in self.rename_dims_map:
        #             self.rename_dims_map[d.key] = d.name
        #         else d

        # if self.dim_name_from_role_name:
        #     d = {v: k for k, v in self.dim_roles.items()}
        #     for k in list(self.rename_dims_map.keys()):
        #         if k in self.dim_roles:
        #             d[self.dim_roles[k]] = self.rename_dims_map.pop(k)

        #     d.update(self.rename_dims_map)
        #     self.rename_dims_map = d

        self.var_key_dim = None

        # ensure all the required keys are in the profile
        keys = []
        for d in self.dims.values():
            keys.append(d.key)

        self.profile.add_keys(keys)

        # self.profile.prepend_keys(self.split_dims)

    def _init_remapping_dims(self, keys):
        # A remapping can be either a variable key or a user specified dimension. They can contain
        # variable related keys, which will be checked later.
        remapping = self.profile.remapping.build()
        remapping_dims = {}
        if remapping:
            for k in remapping.lists:
                d = RemappingDim(self, k, remapping.components(k))
                if k == self.profile.variable.key:
                    self.var_key_dim = d
                elif k in keys:
                    remapping_dims[k] = d
                # else:
                #     raise ValueError(f"Key {k} found in remapping but not in extra_dims or ensure_dims")
        return remapping_dims

    def _init_fixed_dims(self):
        assert self.fixed_dims

        if self.profile.variable.key in self.fixed_dims:
            raise ValueError((f"Variable key {self.profile.variable.key} cannot be in fixed_dims."))
        if self.extra_dims:
            raise ValueError(f"extra_dims={self.extra_dims} cannot be used with fixed_dims")
        if self.drop_dims:
            raise ValueError(f"drop_dims={self.drop_dims} cannot be used with fixed_dims")

        # if self.ensure_dims:
        #     for k in self.ensure_dims:
        #         if k not in self.fixed_dims and k not in self.split_dims:
        #             raise ValueError(
        #                 (
        #                     f'Key "{k}" found in ensure_dims but not found in fixed_dims! '
        #                     "When fixed_dims specified ensure_dims can only contain keys "
        #                     "from fixed_dims"
        #                 )
        #             )

        # self.ensure_dims = [k for k in self.fixed_dims]
        self.ensure_dims = list(self.fixed_dims.keys())
        dims = {k: make_dim(self, name=k, key=v) for k, v in self.fixed_dims.items()}
        return dims

    def _init_dims(self):
        assert not self.fixed_dims

        for k in self.drop_dims:
            if k in self.extra_dims:
                raise ValueError(f"Key {k} cannot be in drop_dims and extra_dims")
            if k in self.ensure_dims:
                raise ValueError(f"Key {k} cannot be in drop_dims and ensure_dims")

        for k in self.split_dims:
            if k not in self.drop_dims and k not in self.extra_dims and k not in self.ensure_dims:
                self.drop_dims.append(k)

        def _remove_duplicates(keys):
            r = []
            for k in keys:
                if k not in r:
                    r.append(k)
            return r

        if self.profile.variable.key:
            var_keys = [self.profile.variable.key]
        else:
            var_keys = []

        # non-core dims
        keys = list(self.extra_dims.keys()) + self.ensure_dims
        keys = _remove_duplicates(keys)

        remapping_dims = self._init_remapping_dims(keys)
        dims = dict(**remapping_dims)

        if not self.var_key_dim:
            assert self.profile.variable.key is not None
            self.var_key_dim = make_dim(self, name=self.profile.variable.key)

        # dims at this point can only contain remapping dims
        for k in keys:
            # note: remapping overrides existing keys
            if k not in dims:
                dims[k] = make_dim(self, name=k)

        dims = {k: v for k, v in dims.items() if k not in self.drop_dims}
        # print(f"dims", dims)

        # initial check for variable-related dimensions
        invalid_var_keys = []
        for k in dims.keys():
            if k in var_keys:
                invalid_var_keys.append(k)

        if invalid_var_keys:
            raise ValueError(self.var_dim_found_error_message(invalid_var_keys))

        # add predefined dimensions
        core_dims = {}
        ignored = {}
        self.core_dim_order = []
        for k, v in DIM_BUILDERS.items():
            builder = v(self.profile, self)
            used, ignored = builder.dims()
            for k, v in used.items():
                core_dims[k] = v
                self.core_dim_order.append(k)

            ignored.update(ignored)

        for k in list(ignored.keys()):
            if k in remapping_dims:
                ignored.pop(k)

        # print(f"ignored", ignored)

        # construct initial dims, ensure core dims are in the right order
        all_dims = {}
        for k in dims:
            if k in remapping_dims or (k not in core_dims and k not in ignored):
                if k in self.ensure_dims or k not in self.drop_dims:
                    all_dims[k] = dims[k]

            # # if k not in self.drop_dims and k not in self.split_dims:
            # if k in self.ensure_dims or k not in self.drop_dims and

            # if k not in self.drop_dims and (k not in self.split_dims or k in self.ensure_dims):
            #     if k in remapping_dims or (k not in core_dims and k not in ignored):
            #         all_dims[k] = dims[k]

        for k in core_dims:
            if k not in all_dims and k not in self.drop_dims and k not in remapping_dims:
                # assert k not in all_dims
                all_dims[k] = core_dims[k]

        # print(f"all_dims", all_dims)
        return all_dims

    def var_dim_found_error_message(self, keys):
        return (
            f'Variable-related keys "{keys}"" cannot be dimensions. Such a key'
            " must be specified as the variable_key. The variable_key can only "
            f'be set to a single key. Its current value is "{self.profile.variable.key}"'
        )

    def register_remapping(self, remapping, patch=None):
        self.profile.remapping.add(remapping, patch)

    def deactivate(self, keys, ignore_dim=None, others=False, collect=False):
        names = []
        for d in self.dims.values():
            if d.active and d != ignore_dim:
                if any(key in d for key in keys):
                    # print(f"deactivate d={d.name} keys={keys}")
                    d.active = False
                    if others:
                        d.deactivate_drop_list()
                    if collect:
                        names.append(d.name)

        if collect:
            return names

    def remove(self, keys, ignore_dim=None, others=False, collect=False):
        self.deactivate(keys, ignore_dim, others, collect)
        self.dims = {k: v for k, v in self.dims.items() if v.active}

    def update(self, ds):
        for d in self.dims.values():
            d.update(ds)

        self.dims = {k: v for k, v in self.dims.items() if v.active}

    @property
    def active_dim_keys(self):
        return [d.key for d in self.dims.values() if d.active]

    def make_coords(self):
        r = {d.coord.name: d.coord.make_var(self.profile) for d in self.dims.values() if d.coord is not None}
        return r

    def as_coord(self, tensor):
        r = {}

        def _get(k):
            for d in self.dims.values():
                if k == d.key:
                    return d

        for k, v in tensor.user_coords.items():
            for d in self.dims.values():
                d = _get(k)
                name, coord = d.as_coord(k, v, tensor.source)
                r[name] = coord

        return r

    def to_list(self):
        return list(self.dims.values())

    def get_dims(self, names):
        r = []
        for name in names:
            if name in self.dims:
                r.append(self.dims[name])
            else:
                r.append(make_dim(self, name=name))
        return r

    def rename_dataset_dims(self, dataset):
        # first rename the dimensions where the name and key are different
        mapping = {}
        for d in self.dims.values():
            if d.key in dataset.dims and d.name != d.key:
                mapping[d.key] = d.name
        if mapping:
            dataset = dataset.rename(mapping)

        # then apply the user defined rename_dims_map
        if self.rename_dims_map:
            mapping = {k: v for k, v in self.rename_dims_map.items() if k in dataset.dims}
            if mapping:
                dataset = dataset.rename(mapping)

        return dataset


PREDEFINED_DIMS = {}
for i, d in enumerate(
    [
        NumberDim,
        ForecastRefTimeDim,
        DateDim,
        TimeDim,
        StepDim,
        ValidTimeDim,
        LevelDim,
        LevelPerTypeDim,
        LevelAndTypeDim,
    ]
):
    if d.name:
        PREDEFINED_DIMS[d.name] = d
    else:
        assert d.alias
        for k in d.alias:
            if k not in PREDEFINED_DIMS:
                PREDEFINED_DIMS[k] = d



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import math
from abc import abstractmethod
from collections import defaultdict
from functools import cached_property

import deprecation
from earthkit.utils.array import array_namespace
from earthkit.utils.array import array_to_numpy
from earthkit.utils.array import convert_array
from earthkit.utils.array import get_backend

from earthkit.data.core import Base
from earthkit.data.core.index import Index
from earthkit.data.core.index import MaskIndex
from earthkit.data.core.index import MultiIndex
from earthkit.data.decorators import cached_method
from earthkit.data.decorators import detect_out_filename
from earthkit.data.utils.metadata.args import metadata_argument


def _bits_per_value_to_metadata(**kwargs):
    # TODO: remove this function when save() and write() are removed
    metadata = {}
    bits_per_value = kwargs.pop("bits_per_value", None)
    if bits_per_value is not None:
        metadata = {"bitsPerValue": bits_per_value}
    return metadata, kwargs


class FieldListIndices:
    def __init__(self, field_list):
        self.fs = field_list
        self.user_indices = dict()

    @cached_property
    def default_index_keys(self):
        if len(self.fs) > 0:
            return self.fs[0]._metadata.index_keys()
        else:
            return []

    def _index_value(self, key):
        values = set()
        for f in self.fs:
            v = f.metadata(key, default=None)
            if v is not None:
                values.add(v)

        return sorted(list(values))

    @cached_property
    def default_indices(self):
        indices = defaultdict(set)
        keys = self.default_index_keys
        for f in self.fs:
            v = f.metadata(keys, default=None)
            for i, k in enumerate(keys):
                if v[i] is not None:
                    indices[k].add(v[i])

        return {k: sorted(list(v)) for k, v in indices.items()}

    def indices(self, squeeze=False):
        r = {**self.default_indices, **self.user_indices}

        if squeeze:
            return {k: v for k, v in r.items() if len(v) > 1}
        else:
            return r

    def index(self, key):
        if key in self.user_indices:
            return self.user_indices[key]

        if key in self.default_index_keys:
            return self.default_indices[key]

        self.user_indices[key] = self._index_value(key)
        return self.user_indices[key]


class Field(Base):
    r"""Represent a Field."""

    @property
    def array_backend(self):
        r""":obj:`ArrayBackend`: Return the array backend of the field."""
        return get_backend(self._values())

    @abstractmethod
    def _values(self, dtype=None):
        r"""Return the raw values extracted from the underlying storage format
        of the field.

        Parameters
        ----------
        dtype: str, array.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``float64``.

        The original shape and array type of the raw values are kept.

        Returns
        -------
        array-like
            Field values.

        """
        self._not_implemented()

    @property
    def values(self):
        r"""array-like: Get the values stored in the field as a 1D array."""
        return self._flatten(self._values())

    @property
    @abstractmethod
    def _metadata(self):
        r"""Metadata: Get the object representing the field's metadata."""
        self._not_implemented()

    def to_numpy(self, flatten=False, dtype=None, index=None):
        r"""Return the values stored in the field as an ndarray.

        Parameters
        ----------
        flatten: bool
            When it is True a flat ndarray is returned. Otherwise an ndarray with the field's
            :obj:`shape` is returned.
        dtype: str, numpy.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is ``float64``.
        index: ndarray indexing object, optional
            The index of the values and to be extracted. When it
            is None all the values are extracted

        Returns
        -------
        ndarray
            Field values

        """
        v = array_to_numpy(self._values(dtype=dtype))
        shape = self._required_shape(flatten)
        if shape != v.shape:
            v = v.reshape(shape)
        if index is not None:
            v = v[index]
        return v

    def to_array(self, flatten=False, dtype=None, array_backend=None, index=None):
        r"""Return the values stored in the field.

        Parameters
        ----------
        flatten: bool
            When it is True a flat array is returned. Otherwise an array with the field's
            :obj:`shape` is returned.
        dtype: str, array.dtype or None
            Typecode or data-type of the array. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is ``float64``.
        array_backend: str, module or None
            The array backend to be used. When it is :obj:`None` the underlying array format
            of the field is used.
        index: array indexing object, optional
            The index of the values and to be extracted. When it
            is None all the values are extracted

        Returns
        -------
        array-array
            Field values.

        """
        v = self._values(dtype=dtype)
        if array_backend is not None:
            v = convert_array(v, target_backend=array_backend)

        v = self._reshape(v, flatten)
        if index is not None:
            v = v[index]
        return v

    def data(self, keys=("lat", "lon", "value"), flatten=False, dtype=None, index=None):
        r"""Return the values and/or the geographical coordinates for each grid point.

        Parameters
        ----------
        keys: :obj:`str`, :obj:`list` or :obj:`tuple`
            Specifies the type of data to be returned. Any combination of "lat", "lon" and "value"
            is allowed here.
        flatten: bool
            When it is True a flat array per key is returned. Otherwise an array with the field's
            :obj:`shape` is returned for each key.
        dtype: str, array.dtype or None
            Typecode or data-type of the arrays. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is ``float64``.
        index: array indexing object, optional
            The index of the values and or the latitudes/longitudes to be extracted. When it
            is None all the values and/or coordinates are extracted.

        Returns
        -------
        array-like
            An multi-dimensional array containing one array per key is returned
            (following the order in ``keys``). The underlying array format
            of the field is used. When ``keys`` is a single value only the
            array belonging to the key is returned.

        Examples
        --------
        - :ref:`/examples/grib_lat_lon_value.ipynb`

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
        >>> d = ds[0].data()
        >>> d.shape
        (3, 7, 12)
        >>> d[0, 0, 0]  # first latitude
        90.0
        >>> d[1, 0, 0]  # first longitude
        0.0
        >>> d[2, 0, 0]  # first value
        272.56417847
        >>> d = ds[0].data(keys="lon")
        >>> d.shape
        (7, 12)
        >>> d[0, 0]  # first longitude
        0.0

        See Also
        --------
        to_latlon
        to_points
        to_numpy
        values

        """
        _keys = dict(
            lat=self._metadata.geography.latitudes,
            lon=self._metadata.geography.longitudes,
            value=self._values,
        )

        if isinstance(keys, str):
            keys = [keys]

        for k in keys:
            if k not in _keys:
                raise ValueError(f"data: invalid argument: {k}")

        r = {}
        for k in keys:
            # TODO: convert dtype
            v = _keys[k](dtype=dtype)
            if v is None:
                raise ValueError(f"data: {k} not available")
            v = self._reshape(v, flatten)
            if index is not None:
                v = v[index]
            r[k] = v

        # convert latlon to array format
        ll = {k: r[k] for k in r if k != "value"}
        if ll:
            sample = r.get("value", None)
            if sample is None:
                sample = self._values(dtype=dtype)
            for k, v in zip(ll.keys(), convert_array(list(ll.values()), target_array_sample=sample)):
                r[k] = v

        r = list(r.values())
        if len(r) == 1:
            return r[0]
        else:
            return array_namespace(r[0]).stack(r)

    def to_points(self, flatten=False, dtype=None, index=None):
        r"""Return the geographical coordinates in the data's original
        Coordinate Reference System (CRS).

        Parameters
        ----------
        flatten: bool
            When it is True 1D arrays are returned. Otherwise arrays with the field's
            :obj:`shape` are returned.
        dtype: str, array.dtype or None
            Typecode or data-type of the arrays. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``float64``.
        index: array indexing object, optional
            The index of the coordinates to be extracted. When it is None
            all the values are extracted.

        Returns
        -------
        dict
            Dictionary with items "x" and "y", containing the arrays of the x and
            y coordinates, respectively. The underlying array format
            of the field is used.

        Raises
        ------
        ValueError
            When the coordinates in the data's original CRS are not available.

        See Also
        --------
        to_latlon

        """
        x = self._metadata.geography.x(dtype=dtype)
        y = self._metadata.geography.y(dtype=dtype)
        r = {}
        if x is not None and y is not None:
            x = self._reshape(x, flatten)
            y = self._reshape(y, flatten)
            if index is not None:
                x = x[index]
                y = y[index]
            r = dict(x=x, y=y)
        elif self.projection().CARTOPY_CRS == "PlateCarree":
            lon, lat = self.data(("lon", "lat"), flatten=flatten, dtype=dtype, index=index)
            return dict(x=lon, y=lat)
        else:
            raise ValueError("to_points(): geographical coordinates in original CRS are not available")

        # convert values to array format
        assert r
        sample = self._values(dtype=dtype)
        for k, v in zip(r.keys(), convert_array(list(r.values()), target_array_sample=sample)):
            r[k] = v
        return r

    def to_latlon(self, flatten=False, dtype=None, index=None):
        r"""Return the latitudes/longitudes of all the gridpoints in the field.

        Parameters
        ----------
        flatten: bool
            When it is True 1D arrays are returned. Otherwise arrays with the field's
            :obj:`shape` are returned.
        dtype: str, array.dtype or None
            Typecode or data-type of the arrays. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``float64``.
        index: array indexing object, optional
            The index of the latitudes/longitudes to be extracted. When it is None
            all the values are extracted.

        Returns
        -------
        dict
            Dictionary with items "lat" and "lon", containing the arrays of the latitudes and
            longitudes, respectively. The underlying array format
            of the field is used.

        See Also
        --------
        to_points

        """
        lon, lat = self.data(("lon", "lat"), flatten=flatten, dtype=dtype, index=index)
        return dict(lat=lat, lon=lon)

    def grid_points(self):
        r = self.to_latlon(flatten=True)
        return r["lat"], r["lon"]

    def grid_points_unrotated(self):
        lat = self._metadata.geography.latitudes_unrotated()
        lon = self._metadata.geography.longitudes_unrotated()
        return lat, lon

    @property
    def rotation(self):
        return self._metadata.geography.rotation

    @property
    def resolution(self):
        return self._metadata.geography.resolution()

    @property
    def mars_grid(self):
        return self._metadata.geography.mars_grid()

    @property
    def mars_area(self):
        return self._metadata.geography.mars_area()

    @property
    def shape(self):
        r"""tuple: Get the shape of the field.

        For structured grids the shape is a tuple in the form of (Nj, Ni) where:

        - ni: the number of gridpoints in i direction (longitude for a regular latitude-longitude grid)
        - nj: the number of gridpoints in j direction (latitude for a regular latitude-longitude grid)

        For other grid types the number of gridpoints is returned as ``(num,)``
        """
        return self._metadata.geography.shape()

    def projection(self):
        r"""Return information about the projection.

        Returns
        -------
        :obj:`Projection`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        return self._metadata.geography.projection()

    def bounding_box(self):
        r"""Return the bounding box of the field.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        return self._metadata.geography.bounding_box()

    def datetime(self):
        r"""Return the date and time of the field.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds[4].datetime()
        {'base_time': datetime.datetime(2020, 12, 21, 12, 0),
        'valid_time': datetime.datetime(2020, 12, 21, 18, 0)}

        """
        return self._metadata.datetime()

    def valid_datetime(self):
        self._not_implemented()

    def base_datetime(self):
        self._not_implemented()

    def h_datetime(self):
        self._not_implemented()

    def an_datetime(self):
        self._not_implemented()

    def indexing_datetime(self):
        self._not_implemented()

    def metadata(self, *keys, astype=None, remapping=None, patches=None, **kwargs):
        r"""Return metadata values from the field.

        When called without any arguments returns a :obj:`Metadata` object.

        Parameters
        ----------
        *keys: tuple
            Positional arguments specifying metadata keys. Can be empty, in this case all
            the keys from the specified ``namespace`` will
            be used. (See examples below).
        astype: type name, :obj:`list` or :obj:`tuple`
            Return types for ``keys``. A single value is accepted and applied to all the ``keys``.
            Otherwise, must have same the number of elements as ``keys``. Only used when
            ``keys`` is not empty.
        remapping: dict, optional
            Creates new metadata keys from existing ones that we can refer to in ``*args`` and
            ``**kwargs``. E.g. to define a new
            key "param_level" as the concatenated value of the "param" and "level" keys use::

                remapping={"param_level": "{param}{level}"}
        **kwargs: dict, optional
            Other keyword arguments:

            * namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
                The namespace to choose the ``keys`` from. When ``keys`` is empty and ``namespace`` is
                :obj:`all` all the available namespaces will be used. When ``keys`` is non empty
                ``namespace`` cannot specify multiple values and it cannot be :obj:`all`. When
                ``namespace`` is None or empty str all the available keys will be used
                (without a namespace qualifier).

            * default: value, optional
                Specifies the same default value for all the ``keys`` specified. When ``default`` is
                **not present** and a key is not found or its value is a missing value
                :obj:`metadata` will raise KeyError.

        Returns
        -------
        single value, :obj:`list`, :obj:`tuple`, :obj:`dict` or :obj:`Metadata`
            - when called without any arguments returns a :obj:`Metadata` object
            - when ``keys`` is not empty:
                - returns single value when ``keys`` is a str
                - otherwise returns the same type as that of ``keys`` (:obj:`list` or :obj:`tuple`)
            - when ``keys`` is empty:
                - when ``namespace`` is None or an empty str returns a :obj:`dict` with all
                  the available keys and values
                - when ``namespace`` is :obj:`str` returns a :obj:`dict` with the keys and values
                  in that namespace
                - otherwise returns a :obj:`dict` with one item per namespace (dict of dict)

        Raises
        ------
        KeyError
            If no ``default`` is set and a key is not found in the message or it has a missing value.

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")

        Calling without arguments:

        >>> r = ds[0].metadata()
        >>> r
        <earthkit.data.readers.grib.metadata.GribMetadata object at 0x164ace170>
        >>> r["name"]
        '2 metre temperature'

        Getting keys with their native type:

        >>> ds[0].metadata("param")
        '2t'
        >>> ds[0].metadata("param", "units")
        ('2t', 'K')
        >>> ds[0].metadata(("param", "units"))
        ('2t', 'K')
        >>> ds[0].metadata(["param", "units"])
        ['2t', 'K']
        >>> ds[0].metadata(["param"])
        ['2t']
        >>> ds[0].metadata("badkey")
        KeyError: 'badkey'
        >>> ds[0].metadata("badkey", default=None)
        <BLANKLINE>

        Prescribing key types:

        >>> ds[0].metadata("centre", astype=int)
        98
        >>> ds[0].metadata(["paramId", "centre"], astype=int)
        [167, 98]
        >>> ds[0].metadata(["centre", "centre"], astype=[int, str])
        [98, 'ecmf']

        Using namespaces:

        >>> ds[0].metadata(namespace="parameter")
        {'centre': 'ecmf', 'paramId': 167, 'units': 'K', 'name': '2 metre temperature', 'shortName': '2t'}
        >>> ds[0].metadata(namespace=["parameter", "vertical"])
        {'parameter': {'centre': 'ecmf', 'paramId': 167, 'units': 'K', 'name': '2 metre temperature',
         'shortName': '2t'},
         'vertical': {'typeOfLevel': 'surface', 'level': 0}}
        >>> r = ds[0].metadata(namespace=all)
        >>> r.keys()
        dict_keys(['default', 'ls', 'geography', 'mars', 'parameter', 'statistics', 'time', 'vertical'])
        >>> r = ds[0].metadata(namespace=None)
        >>> len(r)
        186
        >>> r["name"]
        '2 metre temperature'
        """

        if remapping is not None or patches is not None:
            from earthkit.data.core.order import build_remapping

            remapping = build_remapping(remapping, patches)
            return remapping(self.metadata)(*keys, astype=astype, **kwargs)

        # when called without arguments returns the metadata object
        if len(keys) == 0 and astype is None and not kwargs:
            return self._metadata

        namespace = kwargs.pop("namespace", None)
        key, namespace, astype, key_arg_type = metadata_argument(*keys, namespace=namespace, astype=astype)

        assert isinstance(key, list)
        assert isinstance(namespace, (list, tuple))

        if key:
            assert isinstance(astype, (list, tuple))
            if namespace and namespace[0] != "default":
                key = [namespace[0] + "." + k for k in key]

            raise_on_missing = "default" not in kwargs
            default = kwargs.pop("default", None)

            r = [
                self._metadata.get(
                    k,
                    default=default,
                    astype=kt,
                    raise_on_missing=raise_on_missing,
                    **kwargs,
                )
                for k, kt in zip(key, astype)
            ]

            if key_arg_type is str:
                return r[0]
            elif key_arg_type is tuple:
                return tuple(r)
            else:
                return r
        elif namespace:
            if all in namespace:
                namespace = self._metadata.namespaces()

            r = {ns: self._metadata.as_namespace(ns) for ns in namespace}
            if len(r) == 1:
                return r[namespace[0]]
            else:
                return r
        else:
            return self._metadata.as_namespace(None)

    def dump(self, namespace=all, **kwargs):
        r"""Generate dump with all the metadata keys belonging to ``namespace``.

        In a Jupyter notebook it is represented as a tabbed interface.

        Parameters
        ----------
        namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
            The namespace to dump. The following `namespace` values
            have a special meaning:

            - :obj:`all`: all the available namespaces will be used.
            - None or empty str: all the available keys will be used
                (without a namespace qualifier)

        **kwargs: dict, optional
            Other keyword arguments used for testing only

        Returns
        -------
        NamespaceDump
            Dict-like object with one item per namespace. In a Jupyter notebook represented
            as a tabbed interface to browse the dump contents.

        Examples
        --------
        :ref:`/examples/grib_metadata.ipynb`

        """
        return self._metadata.dump(namespace=namespace, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", details="Use to_target() instead")
    def save(self, filename, append=False, **kwargs):
        r"""Write the field into a file.

        Parameters
        ----------
        filename: str, optional
            The target file path, if not defined attempts will be made to detect the filename
        append: bool, optional
            When it is true append data to the target file. Otherwise
            the target file be overwritten if already exists. Default is False
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`write`.
        """
        metadata = kwargs.pop("metadata", None)
        if metadata is None:
            metadata, kwargs = _bits_per_value_to_metadata(**kwargs)
        self.to_target("file", filename, append=append, metadata=metadata, **kwargs)
        # the original implementation
        # flag = "wb" if not append else "ab"
        # with open(filename, flag) as f:
        #     self.write(f, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", details="Use to_target() instead")
    def write(self, f, **kwargs):
        metadata = kwargs.pop("metadata", None)
        if metadata is None:
            metadata, kwargs = _bits_per_value_to_metadata(**kwargs)
        self.to_target("file", f, metadata=metadata, **kwargs)

    def to_target(self, target, *args, **kwargs):
        r"""Write the field into a target object.

        Parameters
        ----------
        target: object
            The target object to write the field into.
        *args: tuple
            Positional arguments used to specify the target object.
        **kwargs: dict, optional
            Other keyword arguments used to write the field into the target object.
        """
        from earthkit.data.targets import to_target

        to_target(target, *args, data=self, **kwargs)

    def default_encoder(self):
        return self._metadata.data_format()

    def _encode(self, encoder, **kwargs):
        """Double dispatch to the encoder"""
        return encoder._encode_field(self, **kwargs)

    def __getitem__(self, key):
        """Return the value of the metadata ``key``."""
        return self._metadata.get(key)

    @abstractmethod
    def message(self):
        r"""Return a buffer containing the encoded message for message based formats (e.g. GRIB).

        Returns
        -------
        bytes
        """
        self._not_implemented()

    def __repr__(self):
        return "%s(%s,%s,%s,%s,%s,%s)" % (
            self.__class__.__name__,
            self._metadata.get("shortName", None),
            self._metadata.get("levelist", None),
            self._metadata.get("date", None),
            self._metadata.get("time", None),
            self._metadata.get("step", None),
            self._metadata.get("number", None),
        )

    def _attributes(self, names, remapping=None, joiner=None, default=None):
        result = {}
        metadata = self._metadata.get
        if remapping is not None:
            metadata = remapping(metadata, joiner=joiner)

        for name in names:
            result[name] = metadata(name, default=default)
        return result

        # return {name: metadata(name) for name in names}

    @abstractmethod
    def clone(self, *, values=None, metadata=None, **kwargs):
        r"""Create a new :class:`ClonedField` the with updated values and/or metadata.

        Parameters
        ----------
        values: array-like or None
            The values to be stored in the new :class:`ClonedField`. When it is ``None`` the resulting
            :class:`ClonedField` will access the values from the original field.
        metadata: dict, :class:`Metadata` or None
            If it is a dictionary, it is merged with ``**kwargs`` and interpreted in the same way
            as ``**kwargs``. If it is a :class:`Metadata` object, it is used as the new metadata. In
            this case ``**kwargs`` cannot be used.
        **kwargs: dict, optional
            Keys and values to update the metadata with. Metadata values can also be ``callables``
            with the following positional arguments: original_field, key, original_metadata.
            The new :class:`ClonedField` will contain a reference to the original metadata object and
            keys not present in ``kwargs`` will be accessed from the original field.

        Returns
        -------
        :class:`ClonedField`
            The new field with updated values and/or metadata keeping a
            reference to the original field.

        Raises
        ------
        ValueError
            If ``metadata`` is a :class:`Metadata` object and ``**kwargs`` is not empty.

        """
        self.not_implemented()

    def copy(self, *, values=None, flatten=False, dtype=None, array_backend=None, metadata=None):
        r"""Create a new :class:`ArrayField` by copying the values and metadata.

        Parameters
        ----------
        values: array-like or None
            The values to be stored in the new :class:`Field`. When it is ``None`` the values
            extracted from the original field by using :obj:`to_array` with ``flatten``, ``dtype``
            and ``array_backend`` and copied to the new field.
        flatten: bool
            Control the shape of the values when they are extracted from the original field.
            When ``True``, flatten the array, otherwise the field's shape is kept. Only used when
            ``values`` is not provided.
        dtype: str, array.dtype or None
            Control the typecode or data-type of the values when they are extracted from
            the original field. If :obj:`None`, the default type used by the underlying
            data accessor is used. For GRIB it is ``float64``. Only used when  ``values``
            is not provided.
        array_backend: str, module or None
            Control the array backend of the values when they are extracted from
            the original field. If :obj:`None`, the underlying array format
            of the field is used. Only used when ``values`` is not provided.
        metadata: :class:`Metadata` or None
            The metadata to be stored in the new :class:`Field`. When it is :obj:`None`
            a copy of the metadata of the original field is used.

        Returns
        -------
        :class:`ArrayField`
        """
        from earthkit.data.sources.array_list import ArrayField

        if values is None:
            values = self.to_array(
                flatten=flatten,
                dtype=dtype,
                array_backend=array_backend,
            )

        if metadata is None:
            metadata = self._metadata.override()

        return ArrayField(
            values,
            metadata,
        )

    def to_xarray(self, *args, **kwargs):
        """Convert the Field into an Xarray Dataset.

        Parameters
        ----------
        *args: tuple
            Positional arguments passed to :obj:`FieldList.to_xarray`.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`FieldList.to_xarray`.

        Returns
        -------
        Xarray Dataset

        """
        return self._to_fieldlist().to_xarray(*args, **kwargs)

    def ls(self, *args, **kwargs):
        r"""Generate a list like summary using a set of metadata keys.

        Parameters
        ----------
        *args: tuple
            Positional arguments passed to :obj:`FieldList.ls`.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`FieldList.ls`.

        Returns
        -------
        Pandas DataFrame
            DataFrame with one row.

        """
        return self._to_fieldlist().ls(*args, **kwargs)

    def describe(self, *args, **kwargs):
        r"""Generate a summary of the Field."""
        return self._to_fieldlist().describe(*args, **kwargs)

    def _to_fieldlist(self):
        return FieldList.from_fields([self])

    @staticmethod
    def _flatten(v):
        """Flatten the array without copying the data."

        Parameters
        ----------
        v: array-like
            The array to be flattened.

        Returns
        -------
        array-like
            1-D array.
        """
        if len(v.shape) != 1:
            n = math.prod(v.shape)
            n = (n,)
            return array_namespace(v).reshape(v, n)
        return v

    def _reshape(self, v, flatten):
        """Reshape the array to the required shape."""
        shape = self._required_shape(flatten)
        if shape != v.shape:
            v = array_namespace(v).reshape(v, shape)
        return v

    def _required_shape(self, flatten, shape=None):
        """Return the required shape of the array."""
        if shape is None:
            shape = self.shape
        return shape if not flatten else (math.prod(shape),)

    def _array_matches(self, array, flatten=False, dtype=None):
        """Check if the array matches the field and conditions."""
        shape = self._required_shape(flatten)
        return shape == array.shape and (dtype is None or dtype == array.dtype)


class FieldList(Index):
    r"""Represent a list of :obj:`Field` \s."""

    def __init__(self, **kwargs):
        if "array_backend" in kwargs:
            import warnings

            warnings.warn(
                (
                    "array_backend option is not supported any longer in FieldList!"
                    " Use to_fieldlist() instead"
                ),
                DeprecationWarning,
            )
            kwargs.pop("array_backend", None)

        super().__init__(**kwargs)

    def _init_from_mask(self, index):
        pass

    def _init_from_multi(self, index):
        pass

    @staticmethod
    def from_fields(fields):
        r"""Create a :class:`SimpleFieldList`.

        Parameters
        ----------
        fields: iterable
            Iterable of :obj:`Field` objects.

        Returns
        -------
        :class:`SimpleFieldList`

        """
        from earthkit.data.indexing.fieldlist import SimpleFieldList

        return SimpleFieldList([f for f in fields])

    @staticmethod
    def from_numpy(array, metadata):
        return FieldList.from_array(array, metadata)

    @staticmethod
    def from_array(array, metadata):
        r"""Create an :class:`SimpleFieldList`.

        Parameters
        ----------
        array: array-like, list
            The fields' values. When it is a list it must contain one array per field.
        metadata: list, :class:`Metadata`
            The fields' metadata. Must contain one :class:`Metadata` object per field. Or
            it can be a single :class:`Metadata` object when all the fields have the same metadata.

        In the generated :class:`SimpleFieldList`, each field is represented by an array
        storing the field values and a :class:`MetaData` object holding
        the field metadata. The shape and dtype of the array is controlled by the ``kwargs``.
        """
        from earthkit.data.sources.array_list import from_array

        return from_array(array, metadata)

    def ignore(self):
        # When the concrete type is Fieldlist we assume the object was
        # created with Fieldlist() i.e. it is empty. We ignore it from
        # all the merge operations.
        if type(self) is FieldList:
            return True
        else:
            return False

    @cached_property
    def _md_indices(self):
        return FieldListIndices(self)

    def indices(self, squeeze=False):
        r"""Return the unique, sorted values for a set of metadata keys (see below)
        from all the fields. Individual keys can be also queried by :obj:`index`.

        Parameters
        ----------
        squeeze : False
            When True only returns the metadata keys that have more than one values.

        Returns
        -------
        dict
            Unique, sorted metadata values from all the
            :obj:`Field`\ s.

        See Also
        --------
        index

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> ds.indices()
        {'class': ['od'], 'stream': ['oper'], 'levtype': ['pl'], 'type': ['an'],
        'expver': ['0001'], 'date': [20180801], 'time': [1200], 'domain': ['g'],
        'number': [0], 'levelist': [300, 400, 500, 700, 850, 1000],
        'param': ['t', 'u', 'v']}
        >>> ds.indices(squeeze=True)
        {'levelist': [300, 400, 500, 700, 850, 1000], 'param': ['t', 'u', 'v']}
        >>> ds.index("level")
        [300, 400, 500, 700, 850, 1000]

        By default :obj:`indices` uses the keys from the
        "mars" :xref:`eccodes_namespace`. Keys with no valid values are not included. Keys
        that :obj:`index` is called with are automatically added to the original set of keys
        used in :obj:`indices`.

        """
        return self._md_indices.indices(squeeze=squeeze)

    def index(self, key):
        r"""Return the unique, sorted values of the specified metadata ``key`` from all the fields.
        ``key`` will be automatically added to the keys returned by :obj:`indices`.

        Parameters
        ----------
        key : str
            Metadata key.

        Returns
        -------
        list
            Unique, sorted values of ``key`` from all the
            :obj:`Field`\ s.

        See Also
        --------
        index

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> ds.index("level")
        [300, 400, 500, 700, 850, 1000]

        """
        return self._md_indices.index(key)

    def _as_array(self, accessor, **kwargs):
        """Use pre-allocated target array to store the field values."""

        def _vals(f):
            return getattr(f, accessor)(**kwargs) if not is_property else getattr(f, accessor)

        n = len(self)
        if n > 0:
            it = iter(self)
            first = next(it)
            is_property = not callable(getattr(first, accessor, None))
            vals = _vals(first)
            first = None
            ns = array_namespace(vals)
            shape = (n, *vals.shape)
            r = ns.empty(shape, dtype=vals.dtype)
            r[0] = vals
            for i, f in enumerate(it, start=1):
                r[i] = _vals(f)
            return r

    def to_numpy(self, **kwargs):
        r"""Return all the fields' values as an ndarray. It is formed as the array of the
        :obj:`data.core.fieldlist.Field.to_numpy` values per field.

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to :obj:`data.core.fieldlist.Field.to_numpy`

        Returns
        -------
        ndarray
            Array containing the field values.

        See Also
        --------
        to_array
        values
        """
        return self._as_array("to_numpy", **kwargs)

    def to_array(self, **kwargs):
        r"""Return all the fields' values as an array. It is formed as the array of the
        :obj:`data.core.fieldlist.Field.to_array` values per field.

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to :obj:`data.core.fieldlist.Field.to_array`

        Returns
        -------
        array-like
            Array containing the field values.

        See Also
        --------
        values
        to_numpy
        """
        return self._as_array("to_array", **kwargs)

    @property
    def values(self):
        r"""array-like: Get all the fields' values as a 2D array. It is formed as the array of
        :obj:`GribField.values <data.readers.grib.codes.GribField.values>` per field.

        See Also
        --------
        to_array

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> for f in ds:
        ...     print(f.values.shape)
        ...
        (209,)
        (209,)
        >>> v = ds.values
        >>> v.shape
        (2, 209)
        >>> v[0][:3]
        array([262.78027344, 267.44726562, 268.61230469])

        """
        return self._as_array("values")

    def data(
        self,
        keys=("lat", "lon", "value"),
        flatten=False,
        dtype=None,
        index=None,
    ):
        r"""Return the values and/or the geographical coordinates.

        Only works when all the fields have the same grid geometry.

        Parameters
        ----------
        keys: :obj:`str`, :obj:`list` or :obj:`tuple`
            Specifies the type of data to be returned. Any combination of "lat", "lon" and "value"
            is allowed here.
        flatten: bool
            When it is True the "lat", "lon" arrays and the "value" arrays per field
            will all be flattened. Otherwise they will preserve the field's :obj:`shape`.
        dtype: str, array.dtype or None
            Typecode or data-type of the arrays. When it is :obj:`None` the default
            type used by the underlying data accessor is used. For GRIB it is
            ``float64``.
        index: array indexing object, optional
            The index of the values to be extracted from each field. When it is None all the
            values are extracted.

        Returns
        -------
        array-like
            The elements of the array (in the order of the ``keys``) are as follows:

            * the latitudes array from the first field  when "lat" is in ``keys``
            * the longitudes array from the first field when "lon" is in ``keys``
            * a values array per field when "values" is in ``keys``

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry.

        Examples
        --------
        - :ref:`/examples/grib_lat_lon_value.ipynb`

        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test6.grib")
        >>> len(ds)
        6
        >>> d = ds.data()
        >>> d.shape
        (8, 7, 12)
        >>> d[0, 0, 0]  # first latitude
        90.0
        >>> d[1, 0, 0]  # first longitude
        0.0
        >>> d[2:, 0, 0]  # first value per field
        array([272.56417847,  -6.28688049,   7.83348083, 272.53916931,
                -4.89837646,   8.66096497])
        >>> d = ds.data(keys="lon")
        >>> d.shape
        (1, 7, 12)
        >>> d[0, 0, 0]  # first longitude
        0.0

        See Also
        --------
        to_latlon
        to_points
        to_numpy
        values

        """
        if self._is_shared_grid():
            if isinstance(keys, str):
                keys = [keys]

            if "lat" in keys or "lon" in keys:
                latlon = self[0].to_latlon(flatten=flatten, dtype=dtype, index=index)

            r = []
            for k in keys:
                if k == "lat":
                    r.append(latlon["lat"])
                elif k == "lon":
                    r.append(latlon["lon"])
                elif k == "value":
                    r.extend([f.to_array(flatten=flatten, dtype=dtype, index=index) for f in self])
                else:
                    raise ValueError(f"data: invalid argument: {k}")
            return array_namespace(r[0]).stack(r)

        elif len(self) == 0:
            return array_namespace(r[0]).array_ns.stack([])
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def metadata(self, *args, **kwargs):
        r"""Return the metadata values for each field.

        Parameters
        ----------
        *args: tuple
            Positional arguments defining the metadata keys. Passed to
            :obj:`GribField.metadata() <data.readers.grib.codes.GribField.metadata>`
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`GribField.metadata() <data.readers.grib.codes.GribField.metadata>`

        Returns
        -------
        list
            List with one item per :obj:`GribField <data.readers.grib.codes.GribField>`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.metadata("param")
        ['2t', 'msl']
        >>> ds.metadata("param", "units")
        [('2t', 'K'), ('msl', 'Pa')]
        >>> ds.metadata(["param", "units"])
        [['2t', 'K'], ['msl', 'Pa']]

        """
        result = []
        for s in self:
            result.append(s.metadata(*args, **kwargs))
        return result

    def ls(self, n=None, keys=None, extra_keys=None, namespace=None):
        r"""Generate a list like summary using a set of metadata keys.

        Parameters
        ----------
        n: int, None
            The number of :obj:`Field`\ s to be
            listed. None means all the messages, ``n > 0`` means fields from the front, while
            ``n < 0`` means fields from the back of the fieldlist.
        keys: list of str, dict, None
            Metadata keys. If it is None the following default set of keys will be used:  "centre",
            "shortName", "typeOfLevel", "level", "dataDate", "dataTime", "stepRange", "dataType",
            "number", "gridType". To specify a column title for each key in the output use a dict.
        extra_keys: list of str, dict, None
            List of additional keys to ``keys``. To specify a column title for each key in the output
            use a dict.
        namespace: str, None
            The namespace to choose the ``keys`` from. When it is set ``keys`` and
            ``extra_keys`` are omitted.

        Returns
        -------
        Pandas DataFrame
            DataFrame with one row per :obj:`Field`.

        See Also
        --------
        head
        tail

        """
        from earthkit.data.utils.summary import ls

        def _proc(keys, n):
            num = len(self)
            pos = slice(0, num)
            if n is not None:
                pos = slice(0, min(num, n)) if n > 0 else slice(num - min(num, -n), num)
            pos_range = range(pos.start, pos.stop)

            if "namespace" in keys:
                ns = keys.pop("namespace", None)
                for i in pos_range:
                    f = self[i]
                    v = f.metadata(namespace=ns)
                    if len(keys) > 0:
                        v.update(f._attributes(keys))
                    yield (v)
            else:
                for i in pos_range:
                    yield (self[i]._attributes(keys))

        _keys = self._default_ls_keys() if namespace is None else dict(namespace=namespace)
        return ls(_proc, _keys, n=n, keys=keys, extra_keys=extra_keys)

    @cached_method
    def _default_ls_keys(self):
        if len(self) > 0:
            return self[0]._metadata.ls_keys()
        else:
            return []

    def head(self, n=5, **kwargs):
        r"""Generate a list like summary of the first ``n`` :obj:`Field`\ s.
        Same as calling :obj:`ls` with ``n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0) to be printed from the front.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        See Also
        --------
        ls
        tail

        Notes
        -----
        The following calls are equivalent:

        .. code-block:: python

            ds.head()
            ds.head(5)
            ds.head(n=5)
            ds.ls(5)
            ds.ls(n=5)

        """
        if n <= 0:
            raise ValueError("head: n must be > 0")
        return self.ls(n=n, **kwargs)

    def tail(self, n=5, **kwargs):
        r"""Generate a list like summary of the last ``n`` :obj:`Field`\ s.
        Same as calling :obj:`ls` with ``-n``.

        Parameters
        ----------
        n: int, None
            The number of messages (``n`` > 0)  to be printed from the back.
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`ls`.

        Returns
        -------
        Pandas DataFrame
            See  :obj:`ls`.

        See Also
        --------
        head
        ls

        Notes
        -----
        The following calls are equivalent:

        .. code-block:: python

            ds.tail()
            ds.tail(5)
            ds.tail(n=5)
            ds.ls(-5)
            ds.ls(n=-5)

        """
        if n <= 0:
            raise ValueError("n must be > 0")
        return self.ls(n=-n, **kwargs)

    def describe(self, *args, **kwargs):
        r"""Generate a summary of the fieldlist."""
        from earthkit.data.utils.summary import format_describe

        def _proc():
            for f in self:
                yield (f._attributes(self._describe_keys()))

        return format_describe(_proc(), *args, **kwargs)

    @cached_method
    def _describe_keys(self):
        if len(self) > 0:
            return self[0]._metadata.describe_keys()
        else:
            return []

    def datetime(self):
        r"""Return the unique, sorted list of dates and times in the fieldlist.

        Returns
        -------
        dict of datatime.datetime
            Dict with items "base_time" and "valid_time".

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "tests/data/t_time_series.grib")
        >>> ds.datetime()
        {'base_time': [datetime.datetime(2020, 12, 21, 12, 0)],
        'valid_time': [
            datetime.datetime(2020, 12, 21, 12, 0),
            datetime.datetime(2020, 12, 21, 15, 0),
            datetime.datetime(2020, 12, 21, 18, 0),
            datetime.datetime(2020, 12, 21, 21, 0),
            datetime.datetime(2020, 12, 23, 12, 0)]}

        """
        base = set()
        valid = set()
        for s in self:
            d = s.datetime()
            base.add(d["base_time"])
            valid.add(d["valid_time"])
        return {"base_time": sorted(base), "valid_time": sorted(valid)}

    def to_points(self, **kwargs):
        r"""Return the geographical coordinates shared by all the fields in
        the data's original Coordinate Reference System (CRS).

        Parameters
        ----------
        **kwargs: dict, optional
            Keyword arguments passed to
            :obj:`Field.to_points() <data.core.fieldlist.Field.to_points>`

        Returns
        -------
        dict
            Dictionary with items "x" and "y", containing the arrays of the x and
            y coordinates, respectively.

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry.
        """
        if self._is_shared_grid():
            return self[0].to_points(**kwargs)
        elif len(self) == 0:
            return dict(x=None, y=None)
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def to_latlon(self, index=None, **kwargs):
        r"""Return the latitudes/longitudes shared by all the fields.

        Parameters
        ----------
        index: array indexing object, optional
            The index of the latitudes/longitudes to be extracted. When it is None
            all the values are extracted.
        **kwargs: dict, optional
            Keyword arguments passed to
            :meth:`Field.to_latlon() <data.core.fieldlist.Field.to_latlon>`

        Returns
        -------
        dict
            Dictionary with items "lat" and "lon", containing the arrays of the latitudes and
            longitudes, respectively.

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> for f in ds:
        ...     print(f.shape)
        ...
        (11, 19)
        (11, 19)
        >>> r = ds.to_latlon()
        >>> for k, v in r.items():
        ...     print(f"{k}: shape={v.shape}")
        ...
        lat: shape=(11, 19)
        lon: shape=(11, 19)
        >>> r["lon"][:2]
        array([[-27., -23., -19., -15., -11.,  -7.,  -3.,   1.,   5.,   9.,  13.,
         17.,  21.,  25.,  29.,  33.,  37.,  41.,  45.],
        [-27., -23., -19., -15., -11.,  -7.,  -3.,   1.,   5.,   9.,  13.,
         17.,  21.,  25.,  29.,  33.,  37.,  41.,  45.]])

        """
        if self._is_shared_grid():
            return self[0].to_latlon(**kwargs)
        elif len(self) == 0:
            return dict(lat=None, lon=None)
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def projection(self):
        r"""Return the projection information shared by all the fields.

        Returns
        -------
        :obj:`Projection`

        Raises
        ------
        ValueError
            When not all the fields have the same grid geometry

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        if self._is_shared_grid():
            return self[0].projection()
        elif len(self) == 0:
            return None
        else:
            raise ValueError("Fields do not have the same grid geometry")

    def bounding_box(self):
        r"""Return the bounding box for each field.

        Returns
        -------
        list
            List with one :obj:`BoundingBox <data.utils.bbox.BoundingBox>` per
            :obj:`Field`
        """
        return [s.bounding_box() for s in self]

    @cached_method
    def _is_shared_grid(self):
        if len(self) > 0:
            grid = self[0]._metadata.geography._unique_grid_id()
            if grid is not None:
                return all(f._metadata.geography._unique_grid_id() == grid for f in self)
        return False

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    @detect_out_filename
    def save(self, filename, append=False, **kwargs):
        r"""Write all the fields into a file.

        Parameters
        ----------
        filename: str, optional
            The target file path, if not defined attempts will be made to detect the filename
        append: bool, optional
            When it is true append data to the target file. Otherwise
            the target file be overwritten if already exists. Default is False
        **kwargs: dict, optional
            Other keyword arguments passed to :obj:`write`.

        See Also
        --------
        :obj:`write`
        :meth:`GribFieldList.save() <data.readers.grib.index.GribFieldList.save>`
        :meth:`SimpleFieldList.save() <data.indexing.fieldlist.SimpleFieldList.save>`

        """
        metadata, kwargs = _bits_per_value_to_metadata(**kwargs)
        self.to_target("file", filename, append=append, metadata=metadata, **kwargs)

        # original code
        # flag = "wb" if not append else "ab"
        # with open(filename, flag) as f:
        #     self.write(f, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def write(self, f, **kwargs):
        r"""Write all the fields to a file object.

        Parameters
        ----------
        f: file object
            The target file object.
        **kwargs: dict, optional
            Other keyword arguments passed to the underlying field implementation.

        See Also
        --------
        read

        """
        metadata, kwargs = _bits_per_value_to_metadata(**kwargs)
        self.to_target("file", f, metadata=metadata, **kwargs)

        # original code
        # for s in self:
        #     s.write(f, **kwargs)

    def default_encoder(self):
        if len(self) > 0:
            return self[0]._metadata.data_format()

    def _encode(self, encoder, **kwargs):
        """Double dispatch to the encoder"""
        return encoder._encode_fieldlist(self, **kwargs)

    def to_tensor(self, *args, **kwargs):
        from earthkit.data.indexing.tensor import FieldListTensor

        return FieldListTensor.from_fieldlist(self, *args, **kwargs)

    def to_fieldlist(self, array_backend=None, **kwargs):
        r"""Convert to a new :class:`FieldList`.

        Parameters
        ----------
        array_backend: str, module, :obj:`ArrayBackend`
            Specifies the array backend for the generated :class:`FieldList`. The array
            type must be supported by :class:`ArrayBackend`.

        **kwargs: dict, optional
            ``kwargs`` are passed to :obj:`to_array` to
            extract the field values the resulting object will store.

        Returns
        -------
        :class:`SimpleFieldList`
            - a new fieldlist containing :class`ArrayField` fields

        Examples
        --------
        The following example will convert a fieldlist read from a file into a
        :class:`SimpleFieldList` storing single precision field values.

        >>> import numpy as np
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/tuv_pl.grib")
        >>> ds.path
        'docs/examples/tuv_pl.grib'
        >>> r = ds.to_fieldlist(array_backend="numpy", dtype=np.float32)
        >>> r
        SimpleFieldList(fields=18)
        >>> hasattr(r, "path")
        False
        >>> r.to_numpy().dtype
        dtype('float32')

        """
        return self.from_fields([f.copy(array_backend=array_backend, **kwargs) for f in self])

    def cube(self, *args, **kwargs):
        from earthkit.data.indexing.cube import FieldCube

        return FieldCube(self, *args, **kwargs)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskFieldList(*args, **kwargs)

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, FieldList) for _ in sources)
        return MultiFieldList(sources)

    def _cache_diag(self):
        """For testing only"""
        from earthkit.data.utils.diag import metadata_cache_diag

        return metadata_cache_diag(self)


class MaskFieldList(FieldList, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiFieldList(FieldList, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import weakref
from importlib import import_module

import deprecation

from earthkit.data.core import Base
from earthkit.data.core.config import CONFIG
from earthkit.data.decorators import detect_out_filename
from earthkit.data.decorators import locked

LOG = logging.getLogger(__name__)


class ReaderMeta(type(Base), type(os.PathLike)):
    pass


class Reader(Base, os.PathLike, metaclass=ReaderMeta):
    appendable = False  # Set to True if the data can be appended to and existing file
    binary = True

    def __init__(self, source, path):
        LOG.debug("Reader for %s is %s", path, self.__class__.__name__)
        self._source = weakref.ref(source)
        self.path = path
        self.source_filename = self.source.source_filename

    @property
    def source(self):
        return self._source()

    @property
    def filter(self):
        return self.source.filter

    @property
    def parts(self):
        if hasattr(self.source, "parts"):
            return self.source.parts

    @property
    def stream(self):
        if hasattr(self.source, "stream"):
            return self.source.stream
        return False

    @property
    def merger(self):
        return self.source.merger

    def mutate(self):
        # Give a chance to `directory` or `zip` to change the reader
        return self

    def mutate_source(self):
        # The source may ask if it needs to mutate
        return None

    def ignore(self):
        # Used by multi-source
        return False

    def cache_file(self, *args, **kwargs):
        return self.source.cache_file(*args, **kwargs)

    @detect_out_filename
    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def save(self, path, **kwargs):
        self.to_target("file", path, **kwargs)

        # original code
        # mode = "wb" if self.binary else "w"
        # with open(path, mode) as f:
        #     self.write(f, **kwargs)

    @deprecation.deprecated(deprecated_in="0.13.0", removed_in=None, details="Use to_target() instead")
    def write(self, f, **kwargs):
        self.to_target("file", f, **kwargs)

        # original code
        # if not self.appendable:
        #     assert f.tell() == 0
        # mode = "rb" if self.binary else "r"
        # with open(self.path, mode) as g:
        #     while True:
        #         chunk = g.read(1024 * 1024)
        #         if not chunk:
        #             break
        #         f.write(chunk)

    def to_target(self, target, *args, **kwargs):
        from earthkit.data.targets import to_target

        to_target(target, *args, data=self, **kwargs)

    def _encode(self, encoder, **kwargs):
        return encoder._encode(self, **kwargs)

    def default_encoder(self):
        return "internal-pass-through"

    def __fspath__(self):
        return self.path

    def index_content(self):
        LOG.warning(f"index-content(): Ignoring {self.path}")
        return []

    def ranges(self):
        self.source._kw


_READERS = {}


# TODO: Add plugins
@locked
def _readers(method_name):
    if not _READERS:
        here = os.path.dirname(__file__)
        for path in sorted(os.listdir(here)):
            if path[0] in ("_", "."):
                continue

            if path.endswith(".py") or os.path.isdir(os.path.join(here, path)):
                name, _ = os.path.splitext(path)
                try:
                    for method in ["reader", "memory_reader", "stream_reader"]:
                        module = import_module(f".{name}", package=__name__)
                        if hasattr(module, method):
                            _READERS[(name, method)] = getattr(module, method)
                            if hasattr(module, "aliases"):
                                for a in module.aliases:
                                    assert a not in _READERS
                                    _READERS[(a, method_name)] = getattr(module, method)
                except Exception:
                    LOG.exception("Error loading reader %s", name)

    return {k[0]: v for k, v in _READERS.items() if k[1] == method_name}


def _find_reader(method_name, source, path_or_data, **kwargs):
    """Helper function to create a reader.

    Tries all the registered methods stored in _READERS.
    """
    for deeper_check in (False, True):
        # We do two passes, the second one
        # allow the plugin to look deeper in the buffer
        for name, r in _readers(method_name).items():
            reader = r(source, path_or_data, deeper_check=deeper_check, **kwargs)
            if reader is not None:
                return reader.mutate()

    return _unknown(method_name, source, path_or_data, **kwargs)


def _unknown(method_name, source, path_or_data, **kwargs):
    from .unknown import UnknownMemoryReader
    from .unknown import UnknownReader
    from .unknown import UnknownStreamReader

    unknowns = {
        "reader": UnknownReader,
        "stream_reader": UnknownStreamReader,
        "memory_reader": UnknownMemoryReader,
    }
    return unknowns[method_name](source, path_or_data, **kwargs)


def _non_existing(source, path, **kwargs):
    if hasattr(source, "empty_reader"):
        return source.empty_reader(path, **kwargs)


def _empty(source, path, **kwargs):
    if hasattr(source, "empty_reader"):
        return source.empty_reader(path, **kwargs)


def reader(source, path, **kwargs):
    """Create the reader for a file/directory specified by path"""
    assert isinstance(path, str), source

    if hasattr(source, "reader"):
        reader = source.reader
        LOG.debug("Looking for a reader for %s (%s)", path, reader)
        if callable(reader):
            return reader(source, path)
        if isinstance(reader, str):
            return _readers()[reader.replace("-", "_")](source, path, magic=None, deeper_check=False)

        raise TypeError("Provided reader must be a callable or a string, not %s" % type(reader))

    if not os.path.exists(path):
        r = _non_existing(source, path, **kwargs)
        if r is not None:
            return r
        raise FileNotFoundError(f"No such file exists: '{path}'")

    LOG.debug("Reader for %s", path)

    if os.path.isdir(path):
        magic = None
    else:
        if os.path.getsize(path) == 0:
            r = _empty(source, path, **kwargs)
            if r is not None:
                return r
            raise Exception(f"File is empty: '{path}'")

        n_bytes = CONFIG.get("reader-type-check-bytes")
        with open(path, "rb") as f:
            magic = f.read(n_bytes)

    LOG.debug("Looking for a reader for %s (%s)", path, magic)

    return _find_reader(
        "reader",
        source,
        path,
        magic=magic,
        **kwargs,
    )


def memory_reader(source, buffer, **kwargs):
    """Create a reader for data held in a memory buffer"""
    assert isinstance(buffer, (bytes, bytearray)), source
    n_bytes = CONFIG.get("reader-type-check-bytes")
    magic = buffer[: min(n_bytes, len(buffer) - 1)]

    return _find_reader("memory_reader", source, buffer, magic=magic, **kwargs)


def stream_reader(source, stream, memory, **kwargs):
    """Create a reader for a stream"""
    magic = None
    if hasattr(stream, "peek") and callable(stream.peek):
        try:
            n_bytes = CONFIG.get("reader-type-check-bytes")
            magic = stream.peek(n_bytes)
            if len(magic) > n_bytes:
                magic = magic[:n_bytes]
        except Exception:
            pass

    return _find_reader(
        "stream_reader",
        source,
        stream,
        magic=magic,
        memory=memory,
        **kwargs,
    )



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest

from earthkit.data import from_source

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_dims  # noqa: E402


@pytest.fixture
def xr_lod_latlon():
    prototype = {
        "latitudes": [10.0, 0.0, -10.0],
        "longitudes": [20, 40.0],
        "values": [1, 2, 3, 4, 5, 6],
        "valid_datetime": "2018-08-01T09:00:00Z",
    }

    d = [
        {"param": "t", "level": 500, **prototype},
        {"param": "t", "level": 850, **prototype},
        {"param": "u", "level": 500, **prototype},
        {"param": "u", "level": 850, **prototype},
    ]
    ds = from_source("list-of-dicts", d)
    return ds


@pytest.fixture
def xr_lod_nongeo():
    prototype = {
        "values": [1, 2, 3, 4, 5, 6],
        "valid_datetime": "2018-08-01T09:00:00Z",
    }

    d = [
        {"param": "t", "level": 500, **prototype},
        {"param": "t", "level": 850, **prototype},
        {"param": "u", "level": 500, **prototype},
        {"param": "u", "level": 850, **prototype},
    ]
    ds = from_source("list-of-dicts", d)
    return ds


@pytest.fixture
def xr_lod_forecast():
    prototype = {
        "latitudes": [10.0, 0.0, -10.0],
        "longitudes": [20, 40.0],
        "values": [1, 2, 3, 4, 5, 6],
        "base_datetime": "2018-08-01T09:00:00Z",
    }

    d = [
        {"param": "t", "level": 500, "step": 0, **prototype},
        {"param": "t", "level": 500, "step": 6, **prototype},
        {"param": "u", "level": 500, "step": 0, **prototype},
        {"param": "u", "level": 500, "step": 6, **prototype},
    ]
    ds = from_source("list-of-dicts", d)
    return ds


def test_xr_engine_lod_latlon(xr_lod_latlon):
    ds_in = xr_lod_latlon
    ds = ds_in.to_xarray(time_dim_mode="raw")

    assert ds is not None
    assert ds["t"].shape == (2, 3, 2)
    assert ds["u"].shape == (2, 3, 2)
    assert np.allclose(ds["latitude"].values, np.array([10.0, 0.0, -10.0]))
    assert np.allclose(ds["longitude"].values, np.array([20.0, 40.0]))


def test_xr_engine_lod_nongeo(xr_lod_nongeo):
    ds_in = xr_lod_nongeo
    ds = ds_in.to_xarray(time_dim_mode="raw")

    assert ds is not None
    assert ds["t"].shape == (2, 6)
    assert ds["u"].shape == (2, 6)

    ref = np.array(
        [
            [1, 2, 3, 4, 5, 6],
            [1, 2, 3, 4, 5, 6],
        ]
    )
    assert np.allclose(ds["t"].values, ref)
    assert np.allclose(ds["u"].values, ref)


def test_xr_engine_lod_forecast(xr_lod_forecast):
    ds_in = xr_lod_forecast
    ds = ds_in.to_xarray(time_dim_mode="forecast")

    assert ds is not None
    assert ds["t"].shape == (2, 3, 2)
    assert ds["u"].shape == (2, 3, 2)

    dims = {"step": [np.timedelta64(0, "h"), np.timedelta64(6, "h")]}
    compare_dims(ds, dims, order_ref_var="t")

    assert np.allclose(ds["latitude"].values, np.array([10.0, 0.0, -10.0]))
    assert np.allclose(ds["longitude"].values, np.array([20.0, 40.0]))



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pytest

from earthkit.data import from_source


@pytest.mark.long_test
@pytest.mark.download
def test_netcdf_opendap():
    url = "https://psl.noaa.gov/thredds/dodsC/Datasets/noaa.oisst.v2/sst.mnmean.nc"
    ds = from_source("opendap", url)

    assert len(ds) == 494
    assert ds[0:2].metadata("variable") == ["sst", "sst"]
    v = ds[0].values
    assert v.shape == (64800,)
    assert np.isclose(v.mean(), 11.8082780)
    v = ds[0].to_numpy()
    assert v.shape == (180, 360)
    assert np.isclose(v.mean(), 11.8082780)

    x = ds.to_xarray()
    assert len(x.data_vars) == 2
    assert list(x["sst"].dims) == ["time", "lat", "lon"]
    assert x["sst"].shape == (494, 180, 360)


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import os
import sys

import pytest

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from forcings_fixtures import load_forcings_fs  # noqa: E402


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
@pytest.mark.parametrize(
    "params,expected_meta",
    [
        (
            dict(param="sin_longitude", valid_datetime="2020-05-13T18:00:00"),
            [["sin_longitude", "2020-05-13T18:00:00"]],
        ),
        (
            dict(
                param=["sin_longitude", "local_time"],
                valid_datetime=["2020-05-14T06:00:00", "2020-05-13T18:00:00"],
            ),
            [
                ["sin_longitude", "2020-05-13T18:00:00"],
                ["local_time", "2020-05-13T18:00:00"],
                ["sin_longitude", "2020-05-14T06:00:00"],
                ["local_time", "2020-05-14T06:00:00"],
            ],
        ),
        (dict(param="invalidval"), []),
        (dict(INVALIDKEY="sin_logitude"), []),
    ],
)
def test_forcings_sel_single_file_1(input_data, params, expected_meta):
    ds, _ = load_forcings_fs(input_data=input_data)

    g = ds.sel(**params)
    assert len(g) == len(expected_meta)
    if len(expected_meta) > 0:
        keys = list(params.keys())
        assert g.metadata(keys) == expected_meta
    return


@pytest.mark.parametrize("input_data", ["grib", "latlon"])
def test_forcings_sel_single_file_as_dict(input_data):
    ds, _ = load_forcings_fs(input_data=input_data)

    g = ds.sel(
        {
            "param": "sin_longitude",
            "valid_datetime": ["2020-05-14T06:00:00", "2020-05-13T18:00:00"],
        }
    )

    assert len(g) == 2
    assert g.metadata(["param", "valid_datetime"]) == [
        ["sin_longitude", "2020-05-13T18:00:00"],
        ["sin_longitude", "2020-05-14T06:00:00"],
    ]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


class _CustomOrder:
    def __call__(self, x, y):
        r = dict(t=2, u=0, v=1)
        a = r[x]
        b = r[y]
        if a == b:
            return 0
        if a > b:
            return 1
        if a < b:
            return -1


@pytest.mark.parametrize(
    "params,expected_meta",
    [
        (
            "ident",
            dict(
                ident=[
                    "01001",
                    "01152",
                    "01241",
                    "01400",
                    "01415",
                    "02836",
                    "02963",
                    "03953",
                    "11035",
                    "11747",
                ]
            ),
        ),
    ],
)
def test_bufr_order_by_single_file(
    params,
    expected_meta,
):
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    g = f.order_by(params)
    assert len(g) == len(f)

    for k, v in expected_meta.items():
        assert g.metadata(k) == v


@pytest.mark.parametrize(
    "params,expected_meta",
    [
        (
            "ident",
            dict(
                ident=[
                    "01001",
                    "01152",
                    "01400",
                    "01415",
                    "02836",
                    "30823",
                    "30846",
                    "60545",
                    "89514",
                    "91648",
                ]
            ),
        ),
        (
            ["ident"],
            dict(
                ident=[
                    "01001",
                    "01152",
                    "01400",
                    "01415",
                    "02836",
                    "30823",
                    "30846",
                    "60545",
                    "89514",
                    "91648",
                ]
            ),
        ),
        (
            ["dataCategory", "ident"],
            dict(
                dataCategory=[0, 0, 0, 0, 0, 2, 2, 2, 2, 2],
                ident=[
                    "30823",
                    "30846",
                    "60545",
                    "89514",
                    "91648",
                    "01001",
                    "01152",
                    "01400",
                    "01415",
                    "02836",
                ],
            ),
        ),
        (
            dict(ident="ascending"),
            dict(
                ident=[
                    "01001",
                    "01152",
                    "01400",
                    "01415",
                    "02836",
                    "30823",
                    "30846",
                    "60545",
                    "89514",
                    "91648",
                ]
            ),
        ),
        (
            dict(ident="descending"),
            dict(
                ident=[
                    "91648",
                    "89514",
                    "60545",
                    "30846",
                    "30823",
                    "02836",
                    "01415",
                    "01400",
                    "01152",
                    "01001",
                ]
            ),
        ),
    ],
)
def test_bufr_order_by_multi_file(params, expected_meta):
    f1 = from_source("file", earthkit_examples_file("temp_10.bufr"))[:5]
    f2 = from_source("file", earthkit_examples_file("synop_10.bufr"))[:5]
    f = from_source("multi", [f1, f2])

    g = f.order_by(params)
    assert len(g) == len(f)

    for k, v in expected_meta.items():
        assert g.metadata(k) == v


def test_bufr_order_by_with_sel():
    f = from_source("file", earthkit_examples_file("temp_10.bufr"))

    g = f.sel(ident=["01400", "01001", "11035"])
    assert len(g) == 3
    r = g.order_by("ident")
    assert len(r) == len(g)
    assert r.metadata("ident") == ["01001", "01400", "11035"]

    g = f.sel(ident=["01400", "01001", "11035"])
    assert len(g) == 3
    r = g.order_by({"ident": "descending"})
    assert len(r) == len(g)
    assert r.metadata("ident") == ["11035", "01400", "01001"]


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math

from earthkit.utils.array import array_namespace

from earthkit.data.core.fieldlist import Field
from earthkit.data.indexing.fieldlist import ClonedFieldCore

LOG = logging.getLogger(__name__)


class ArrayField(Field):
    r"""Represent a field consisting of an array and metadata object.

    Parameters
    ----------
    array: array
        Array storing the values of the field
    metadata: :class:`Metadata`
        Metadata object describing the field metadata.
    """

    def __init__(self, array, metadata):
        if isinstance(array, list):
            import numpy as np

            array = np.array(array)

        if isinstance(metadata, dict):
            from earthkit.data.utils.metadata.dict import UserMetadata

            metadata = UserMetadata(metadata, shape=array.shape)

        # TODO: this solution is questionable due to performance reasons
        if metadata is not None:
            metadata = metadata._hide_internal_keys()

        self._metadata_ = metadata
        self._array_ = array

    @property
    def _array(self):
        return self._array_

    def _values(self, dtype=None):
        """Native array type"""
        if dtype is None:
            return self._array
        else:
            return array_namespace(self._array).astype(self._array, dtype, copy=False)

    @property
    def shape(self):
        v = super().shape
        return v if v is not None else self._array.shape

    def __repr__(self):
        return self.__class__.__name__ + "(%s,%s,%s,%s,%s,%s)" % (
            self._metadata.get("shortName", None),
            self._metadata.get("levelist", None),
            self._metadata.get("date", None),
            self._metadata.get("time", None),
            self._metadata.get("step", None),
            self._metadata.get("number", None),
        )

    def _encode(self, encoder, **kwargs):
        """Double dispatch to the encoder"""
        values = kwargs.pop("values", None)
        if values is None:
            values = self.to_numpy(flatten=True)

        return encoder._encode_field(self, values=values, **kwargs)

    @property
    def _metadata(self):
        return self._metadata_

    @property
    def handle(self):
        return self._metadata._handle

    def _release(self):
        self._array_ = None
        self._metadata_ = None

    def __getstate__(self) -> dict:
        ret = {}
        ret["_array"] = self._array
        ret["_metadata"] = self._metadata
        return ret

    def __setstate__(self, state: dict):
        self._array_ = state.pop("_array")
        self._metadata_ = state.pop("_metadata")

    def clone(self, **kwargs):
        return ClonedArrayField(self, **kwargs)


class ClonedArrayField(ClonedFieldCore, ArrayField):
    def __init__(self, field, **kwargs):
        ClonedFieldCore.__init__(self, field, **kwargs)
        ArrayField.__init__(self, field._array, None)


def from_array(array, metadata):
    def _shape_match(shape1, shape2):
        if shape1 == shape2:
            return True
        if len(shape1) == 1 and shape1[0] == math.prod(shape2):
            return True
        return False

    if not isinstance(metadata, list):
        metadata = [metadata]

    # array_ns = get_backend(self._array).array_ns
    if isinstance(array, list):
        if len(array) == 0:
            raise ValueError("array must not be empty")

    if not isinstance(array, list):
        array_ns = array_namespace(array)
        if array_ns is None:
            raise ValueError(f"array type {type(array)} is not supported")
        elif array.shape[0] != len(metadata):
            # we have a single array and a single metadata
            if len(metadata) == 1 and _shape_match(array.shape, metadata[0].geography.shape()):
                array = array_ns.stack([array])
            else:
                raise ValueError(
                    (
                        f"first array dimension={array.shape[0]} differs "
                        f"from number of metadata objects={len(metadata)}"
                    )
                )
    else:
        if len(array) != len(metadata):
            raise ValueError(
                (f"array len=({len(array)}) differs " f"from number of metadata objects=({len(metadata)})")
            )

    fields = []
    for i, a in enumerate(array):
        if len(metadata) == 1:
            fields.append(ArrayField(a, metadata[0]))
        else:
            fields.append(ArrayField(a, metadata[i]))

    from earthkit.data.indexing.fieldlist import SimpleFieldList

    return SimpleFieldList(fields)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import os
import sys

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.utils.xarray.profile import Profile

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_dim_order  # noqa: E402
from xr_engine_fixtures import load_wrapped_fieldlist  # noqa: E402

DS_LEV = {
    "class": ["od"],
    "param": ["t", "r"],
    "levelist": [1000, 850],
    "level": [1000, 850],
    "time": ["12"],
    "step": [0],
}

DS_DATE_LEV = {
    "class": ["od"],
    "param": ["t", "r"],
    "shortName": ["t", "r"],
    "levelist": [1000, 850],
    "level": [1000, 850],
    "levtype": ["pl"],
    "date": ["20210101", "20210102"],
    "time": ["12"],
    "step": [0],
    # "base_datetime": ["2021-01-01T12:00:00", "2021-01-02T12:00:00"],
}

DS_DATE_STEPS_LEVEL = {
    "class": ["od"],
    "param": ["t", "r"],
    "shortName": ["t", "r"],
    "levelist": [1000, 850],
    "level": [1000, 850],
    "levtype": ["pl"],
    "date": ["20210101", "20210102"],
    "time": ["12"],
    "step": [0, 6],
    # "base_datetime": ["2021-01-01T12:00:00", "2021-01-02T12:00:00"],
}

DS_DATE_SFC_PL = [
    {
        "class": ["od"],
        "param": ["2t", "msl"],
        "shortName": ["2t", "msl"],
        "levelist": [None],
        "level": [None],
        "levtype": ["sfc"],
        "date": ["20210101", "20210102"],
        "time": ["12"],
        "step": [0],
    },
    {
        "class": ["od"],
        "param": ["t", "r"],
        "shortName": ["t", "r"],
        "levelist": [1000, 850],
        "level": [1000, 850],
        "levtype": ["pl"],
        "date": ["20210101", "20210102"],
        "time": ["12"],
        "step": [0],
    },
]


def _attributes(ds):
    return {k: v[0] for k, v in ds.indices().items() if len(v) == 1}


def test_xr_dims_input_fieldlist():
    prof = Profile.make("mars")
    ds = load_wrapped_fieldlist(DS_DATE_LEV, prof)
    assert ds.index("param") == ["r", "t"]
    assert ds.index("param_level") == ["r1000", "r850", "t1000", "t850"]

    remapping = {"param_level": "{param}_{level}"}
    prof = Profile.make("mars", variable_key="param_level", remapping=remapping)
    ds = load_wrapped_fieldlist(DS_DATE_LEV, prof, remapping=prof.remapping.build())
    assert ds.index("param") == ["r", "t"]
    assert ds.index("param_level") == ["r_1000", "r_850", "t_1000", "t_850"]

    remapping = {
        "_class": "_{class}",
        "level_and_type": "{level}_{levtype}",
    }
    prof = Profile.make(
        "mars", variable_key="param", remapping=remapping, extra_dims=["_class", "level_and_type"]
    )
    ds = load_wrapped_fieldlist(DS_DATE_LEV, prof, remapping=prof.remapping.build())
    assert ds.index("param") == ["r", "t"]
    assert ds.index("_class") == ["_od"]
    assert ds.index("level_and_type") == ["1000_pl", "850_pl"]


@pytest.mark.parametrize(
    "kwargs,var_key,variables,dim_keys",
    [
        ({}, "param", ["r", "t"], ["step_timedelta", "levelist"]),
        (
            {"time_dim_mode": "forecast", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            ["step_timedelta", "levelist"],
        ),
        (
            {"squeeze": False, "time_dim_mode": "raw", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            ["time", "step_timedelta", "levelist"],
        ),
    ],
)
def test_xr_dims_ds_lev(kwargs, var_key, variables, dim_keys):
    """Test for the internal profile/dimension object. Cannot use all the options since
    many tasks are performed elsewhere in the engine."""
    # TODO: consider removing this test
    prof = Profile.make("mars", **kwargs)
    ds = load_wrapped_fieldlist(DS_LEV, prof)
    # prof.update(ds, _attributes(ds))
    prof.update(ds)
    assert prof.variable.key == var_key
    assert prof.variable.variables == variables
    assert prof.dim_keys == dim_keys


@pytest.mark.parametrize(
    "kwargs,var_key,variables,dims",
    [
        (
            {"time_dim_mode": "forecast", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            ["forecast_reference_time", "step_timedelta", "levelist", "levtype"],
        ),
        (
            {"time_dim_mode": "raw", "variable_key": "param_level", "dim_name_from_role_name": False},
            "param_level",
            ["r1000", "r850", "t1000", "t850"],
            ["date", "time", "step_timedelta", "levtype"],
        ),
        (
            {
                "time_dim_mode": "raw",
                "variable_key": "param_level",
                "remapping": {"param_level": "{param}_{level}"},
                "dim_name_from_role_name": False,
            },
            "param_level",
            ["r_1000", "r_850", "t_1000", "t_850"],
            ["date", "time", "step_timedelta", "levtype"],
        ),
        (
            {"time_dim_mode": "raw", "variable_key": "shortName", "dim_name_from_role_name": False},
            "shortName",
            ["r", "t"],
            ["date", "time", "step_timedelta", "levelist", "levtype"],
        ),
        (
            {
                "time_dim_mode": "raw",
                "variable_key": "shortName",
                "drop_variables": ["r"],
                "dim_name_from_role_name": False,
            },
            "shortName",
            ["t"],
            ["date", "time", "step_timedelta", "levelist", "levtype"],
        ),
        (
            {
                "time_dim_mode": "raw",
                "variable_key": "param_level",
                "drop_variables": ["r", "r1000"],
                "dim_name_from_role_name": False,
            },
            "param_level",
            ["r850", "t1000", "t850"],
            [
                "date",
                "time",
                "step_timedelta",
                "levtype",
            ],
        ),
        (
            {"time_dim_mode": "raw", "level_dim_mode": "level_and_type", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            {
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "level_and_type": ["1000pl", "850pl"],
            },
        ),
        (
            {"time_dim_mode": "raw", "extra_dims": "class", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            {
                "class": ["od"],
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "levelist": [850, 1000],
                "levtype": ["pl"],
            },
        ),
        (
            {"time_dim_mode": "raw", "ensure_dims": "class", "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            {
                "class": ["od"],
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "levelist": [850, 1000],
                "levtype": ["pl"],
            },
        ),
        (
            {"time_dim_mode": "raw", "ensure_dims": ["class", "step"], "dim_name_from_role_name": False},
            "param",
            ["r", "t"],
            {
                "class": ["od"],
                "step": [0],
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "levelist": [850, 1000],
                "levtype": ["pl"],
            },
        ),
        (
            {
                "time_dim_mode": "raw",
                "ensure_dims": ["class", "step_timedelta"],
                "dim_name_from_role_name": False,
            },
            "param",
            ["r", "t"],
            {
                "class": ["od"],
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "levelist": [850, 1000],
                "levtype": ["pl"],
            },
        ),
        (
            {
                "time_dim_mode": "raw",
                "extra_dims": "class",
                "squeeze": False,
                "dim_name_from_role_name": False,
            },
            "param",
            ["r", "t"],
            {
                "class": ["od"],
                "date": ["20210101", "20210102"],
                "time": ["12"],
                "step_timedelta": [datetime.timedelta(hours=0)],
                "levelist": [850, 1000],
                "levtype": ["pl"],
            },
        ),
    ],
)
def test_xr_dims_ds_date_lev(kwargs, var_key, variables, dims):
    """Test for the internal profile/dimension object. Cannot use all the options since
    many tasks are performed elsewhere in the engine."""
    # TODO: consider removing this test

    prof = Profile.make("mars", **kwargs)
    ds = load_wrapped_fieldlist(DS_DATE_LEV, prof, remapping=prof.remapping.build())

    # print("remapping", prof.remapping.build())
    # print(f"ds: {ds.indices()}")

    # ds.load(prof.index_keys)
    # prof.update(ds, _attributes(ds))
    prof.update(ds)
    assert prof.variable.key == var_key
    assert prof.variable.variables == variables

    if isinstance(dims, list):
        assert prof.dim_keys == dims
    elif isinstance(dims, dict):
        assert prof.dim_keys == list(dims.keys())
        for k in prof.dim_keys:
            assert ds.index(k) == dims[k]
    else:
        raise ValueError(f"Unsupported dims: {dims}")


@pytest.mark.parametrize(
    "kwargs,var_key,variables,dim_keys",
    [
        (
            {"time_dim_mode": "raw"},
            "param",
            ["2t", "msl", "r", "t"],
            ["date", "time", "step_timedelta", "levelist", "levtype"],
        ),
        # ({"base_datetime_dim": True}, "param", ["r", "t"], ["levelist", "levtype"]),
        # ({"squeeze": False}, "param", ["r", "t"], ["time", "step", "levelist", "levtype"]),
    ],
)
def test_xr_dims_ds_sfc_and_pl(kwargs, var_key, variables, dim_keys):
    """Test for the internal profile/dimension object. Cannot use all the options since
    many tasks are performed elsewhere in the engine."""
    # TODO: consider removing this test
    prof = Profile.make("mars", **kwargs)
    ds = load_wrapped_fieldlist(DS_DATE_SFC_PL, prof)
    # prof.update(ds, _attributes(ds))
    prof.update(ds)
    assert prof.variable.key == var_key
    assert prof.variable.variables == variables
    assert prof.dim_keys == dim_keys


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dim_keys",
    [
        (
            {
                "profile": "mars",
                "time_dim_mode": "raw",
                "rename_dims": {"levelist": "zz"},
                "dim_name_from_role_name": False,
            },
            ["date", "time", "step_timedelta", "zz"],
        ),
        (
            {
                "profile": "mars",
                "time_dim_mode": "raw",
                "rename_dims": {"level": "zz"},
            },
            ["date", "time", "step", "zz"],
        ),
    ],
)
def test_xr_rename_dims(kwargs, dim_keys):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds = ds_ek.to_xarray(**kwargs)
    num = len(ds)

    dim_keys = dim_keys + ["latitude", "longitude"]
    assert len(ds) == num

    for v in ds:
        compare_dim_order(ds, dim_keys, v)


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dim_keys",
    [
        (
            {
                "profile": "mars",
                "fixed_dims": ["date", "time", "step", "level"],
            },
            ["date", "time", "step", "level"],
        ),
        (
            {
                "profile": "mars",
                "fixed_dims": ["level", "date", "time", "step"],
            },
            ["level", "date", "time", "step"],
        ),
        (
            {
                "profile": "mars",
                "fixed_dims": [{"my_date": "date"}, ("my_time", "time"), "step", "level"],
            },
            ["my_date", "my_time", "step", "level"],
        ),
        (
            {
                "profile": "mars",
                "fixed_dims": ["forecast_reference_time", "endStep", "level"],
            },
            ["forecast_reference_time", "endStep", "level"],
        ),
        (
            {
                "profile": "mars",
                "fixed_dims": ["forecast_reference_time", ("step", "endStep"), "level"],
            },
            ["forecast_reference_time", "step", "level"],
        ),
    ],
)
def test_xr_fixed_dims(kwargs, dim_keys):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds = ds_ek.to_xarray(**kwargs)
    num = len(ds)

    dim_keys = dim_keys + ["latitude", "longitude"]
    assert len(ds) == num

    for v in ds:
        compare_dim_order(ds, dim_keys, v)


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs,dim_keys",
    [
        (
            {
                "profile": "mars",
                "drop_dims": "number",
                "time_dim_mode": "raw",
                "squeeze": False,
                "dim_name_from_role_name": True,
            },
            ["date", "time", "step", "level", "level_type"],
        ),
        (
            {
                "profile": "mars",
                "drop_dims": ["level_type", "number"],
                "time_dim_mode": "raw",
                "squeeze": False,
                "dim_name_from_role_name": True,
            },
            ["date", "time", "step", "level"],
        ),
        (
            {
                "profile": "mars",
                "drop_dims": "number",
                "time_dim_mode": "raw",
                "squeeze": False,
                "dim_name_from_role_name": False,
            },
            ["date", "time", "step_timedelta", "levelist", "levtype"],
        ),
        (
            {
                "profile": "mars",
                "drop_dims": ["levtype", "number"],
                "time_dim_mode": "raw",
                "squeeze": False,
                "dim_name_from_role_name": False,
            },
            ["date", "time", "step_timedelta", "levelist"],
        ),
    ],
)
def test_xr_drop_dims(kwargs, dim_keys):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds = ds_ek.to_xarray(**kwargs)
    num = len(ds)

    dim_keys = dim_keys + ["latitude", "longitude"]
    assert len(ds) == num

    for v in ds:
        compare_dim_order(ds, dim_keys, v)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from . import Reader
from .netcdf.fieldlist import XArrayFieldList


class ZarrReader(XArrayFieldList, Reader):

    def __init__(self, source, path, **kwargs):
        Reader.__init__(self, source, path, **kwargs)
        XArrayFieldList.__init__(self, self._open_zarr(**kwargs))

    def mutate_source(self):
        return self

    def to_xarray(self, **kwargs):
        return self._open_zarr(**kwargs)

    def _open_zarr(self, **kwargs):
        import xarray as xr

        options = kwargs.get("xarray_open_zarr_kwargs", kwargs)
        return xr.open_zarr(self.path, **options)

    def __repr__(self):
        return f"ZarrReader({self.path})"


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    if (
        os.path.exists(os.path.join(path, ".zarray"))
        or os.path.exists(os.path.join(path, ".zgroup"))
        or os.path.exists(os.path.join(path, ".zmetadata"))
        or os.path.exists(os.path.join(path, ".zattrs"))
    ):
        return ZarrReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from abc import ABCMeta
from abc import abstractmethod
from functools import lru_cache
from importlib import import_module

LOG = logging.getLogger(__name__)


class Target(metaclass=ABCMeta):
    """
    Represent a target.

    Parameters:
    -----------
    encoder: str, Encoder, None
        The encoder to use to encode the data. Can be overridden in the the :obj:`write` method.
        When a string is passed, the encoder is looked up in the available encoders. When None,
        the encoder will be determined from the data to write (if possible) or from
        the :class:`Target` properties.
    template: obj, None
        The template to use to encode the data. Can be overridden in the :obj:`write` method.
    metadata: dict, None
        Metadata to pass to the encoder.

    The :class:`Target` is used to write data to a specific location. The target can be
    a file, a database, a remote server, etc.
    """

    def __init__(self, encoder=None, template=None, metadata=None, **kwargs):
        self._encoder = encoder
        self._template = template
        self._metadata = metadata or {}
        self._closed = False

    @abstractmethod
    def write(
        self,
        data=None,
        encoder=None,
        template=None,
        metadata=None,
        **kwargs,
    ):
        """
        Write data to the target using the given encoder.

        Parameters:
        -----------
        data: obj, None
            The data object to write. If None, the encoder will use all the other arguments
            to generate the data to write.
        encoder: str, Encoder, None
            The encoder to use to encode the data.
            When a string is passed, the encoder is looked up in the available encoders. When None,
            the encoder the :class:`Target` was created with will be used if available. Otherwise,
            the encoder will be determined from the data to write (if possible) or from
            the :class:`Target` properties.
        template: obj, None
            The template to use to encode the data. When None,
            the template the :class:`Target` was created with will be used if available.
        metadata: dict, None
            Metadata to pass to the encoder.
        **kwargs: dict
            Other keyword arguments passed to the encoder.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        pass

    @abstractmethod
    def close(self):
        """Close the target.

        The implementation must close the target and release any resources.
        It must also call :obj:`_mark_closed`. The target will not be able
        to write anymore.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        pass

    @abstractmethod
    def flush(self):
        """Flush the target.

        Some targets may require to flush the data to the underlying storage.

        Raises:
        -------
        ValueError: If the target is already closed.
        """
        pass

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        pass

    @property
    def closed(self):
        """True if the target is closed, False otherwise."""
        return self._closed

    def _mark_closed(self):
        """Register the target as closed."""
        self._raise_if_closed()
        self._closed = True

    def _raise_if_closed(self):
        """Raise an error if the target is closed."""
        if self.closed:
            raise ValueError("Target is closed")


class SimpleTarget(Target):
    def write(
        self,
        data=None,
        **kwargs,
    ):
        self._raise_if_closed()

        if data is not None:
            if hasattr(data, "sources"):
                for d in data.sources:
                    self.write(d, **kwargs)

            self._write(data, **kwargs)
        else:
            self._write(None, **kwargs)

    @abstractmethod
    def _write(self, data, **kwargs):
        """Write generic data to the target.

        Parameters:
        -----------
        data:
            Data to write to the target.
        """
        pass

    def _encode(self, data, encoder=None, template=None, suffix=None, **kwargs):
        """Helper method to encode data.

        Returns
        -------
        :class:`EncodedData`
            The encoded data.
        """
        from earthkit.data.encoders import make_encoder

        if encoder is None:
            encoder = self._encoder

        encoder = make_encoder(data, encoder, suffix=suffix, metadata=self._metadata)

        if template is None:
            template = self._template

        return encoder.encode(data, template=template, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class TargetLoader:
    kind = "target"

    def load_module(self, module):
        return import_module(module, package=__name__).target

    def load_entry(self, entry):
        entry = entry.load()
        if callable(entry):
            return entry
        return entry.target

    def load_remote(self, name):
        return None


class TargetMaker:
    TARGETS = {}

    def __call__(self, name_or_target, *args, **kwargs):
        if isinstance(name_or_target, Target):
            return name_or_target

        name = name_or_target

        loader = TargetLoader()

        if name in self.TARGETS:
            klass = self.TARGETS[name]
        else:
            from earthkit.data.core.plugins import find_plugin

            klass = find_plugin(os.path.dirname(__file__), name, loader)
            self.TARGETS[name] = klass

        target = klass(*args, **kwargs)

        if getattr(target, "name", None) is None:
            target.name = name

        return target

    def __getattr__(self, name):
        return self(name.replace("_", "-"))


create_target = TargetMaker()


@lru_cache
def target_kwargs(target_type):
    import inspect

    r = inspect.signature(target_type.__init__)
    v = []
    for p in r.parameters.values():
        if p.kind == p.KEYWORD_ONLY or p.kind == (p.POSITIONAL_OR_KEYWORD and p.default is not p.empty):
            v.append(p.name)
    return v


def to_target(target, *args, **kwargs):
    """Write data to a target.

    This is a top level function that writes data to a target.

    Parameters:
    -----------
    target: str
        The target to write to. Must be a string.
    """

    # data = kwargs.pop("data", None)

    with create_target(target, *args, **kwargs) as t:
        for k in [*target_kwargs(type(t)), *target_kwargs(Target)]:
            kwargs.pop(k, None)

        t.write(**kwargs)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import pytest

from earthkit.data import from_source
from earthkit.data.testing import NO_COVJSONKIT
from earthkit.data.testing import earthkit_test_data_file


def test_covjson():
    ds = from_source("file", earthkit_test_data_file("time_series.covjson"))
    assert ds


@pytest.mark.skipif(NO_COVJSONKIT, reason="no covjsonkit available")
def test_covjson_to_xarray():
    ds = from_source("file", earthkit_test_data_file("time_series.covjson"))
    assert ds
    a = ds.to_xarray()
    assert len(a.data_vars) == 1


@pytest.mark.skipif(NO_COVJSONKIT, reason="no covjsonkit available")
def test_covjson_memory():
    with open(earthkit_test_data_file("time_series.covjson"), "r") as f:
        d = f.read().encode()

    ds = from_source("memory", d)
    assert ds
    a = ds.to_xarray()
    assert len(a.data_vars) == 1


@pytest.mark.skipif(NO_COVJSONKIT, reason="no covjsonkit available")
def test_covjson_stream():
    stream = open(earthkit_test_data_file("time_series.covjson"), "rb")

    ds = from_source("stream", stream)
    assert ds
    it = iter(ds)
    c = next(it)
    a = c.to_xarray()
    assert len(a.data_vars) == 1

    with pytest.raises(StopIteration):
        next(it)


@pytest.mark.skipif(NO_COVJSONKIT, reason="no covjsonkit available")
def test_covjson_stream_memory():
    stream = open(earthkit_test_data_file("time_series.covjson"), "rb")

    ds = from_source("stream", stream, read_all=True)
    assert ds
    a = ds.to_xarray()
    assert len(a.data_vars) == 1


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


# from emohawk.metadata import AXES, COMPONENTS

# from earthkit.data.readers import netcdf
from earthkit.data.wrappers import Wrapper


class XArrayDataArrayWrapper(Wrapper):
    """Wrapper around an xarray `DataArray`, offering polymorphism and
    convenience methods.
    """

    def __init__(self, data):
        self.data = data

    # def axis(self, axis):
    #     """
    #     Get the data along a specific coordinate axis.

    #     Parameters
    #     ----------
    #     axis : str
    #         The coordinate axis along which to extract data. Accepts values of
    #         `x`, `y`, `z` (vertical level) or `t` (time) (case-insensitive).

    #     Returns
    #     -------
    #     xarray.core.dataarray.DataArray
    #         An xarray `DataArray` containing the data along the given
    #         coordinate axis.
    #     """
    #     for coord in self.source.coords:
    #         if self.source.coords[coord].attrs.get("axis", "").lower() == axis:
    #             break
    #     else:
    #         candidates = AXES.get(axis, [])
    #         for coord in candidates:
    #             if coord in self.source.coords:
    #                 break
    #         else:
    #             raise ValueError(f"No coordinate found with axis '{axis}'")
    #     return self.source.coords[coord]

    def to_xarray(self, *args, **kwargs):
        """Return an xarray representation of the data.

        Returns
        -------
        xarray.core.dataarray.DataArray
        """
        return self.data

    def to_numpy(self, flatten=False):
        """Return a numpy `ndarray` representation of the data.

        Returns
        -------
        numpy.ndarray
        """
        arr = self.data.to_numpy()
        if flatten:
            arr = arr.flatten()
        return arr

    def to_pandas(self, *args, **kwargs):
        """Return a pandas `dataframe` representation of the data.

        Returns
        -------
        pandas.core.frame.DataFrame
        """
        return self.data.to_dataframe(*args, **kwargs)

    def to_netcdf(self, *args, **kwargs):
        """Save the data to a netCDF file.

        Parameters
        ----------
        See `xarray.DataArray.to_netcdf`.
        """
        return self.data.to_netcdf(*args, **kwargs)

    def _encode(self, encoder, **kwargs):
        """Encode the data using the specified encoder.

        Parameters
        ----------
        encoder : Encoder
            The encoder to use for encoding the data.
        **kwargs : dict
            Additional keyword arguments to pass to the encoder.

        Returns
        -------
        EncodedData
            The encoded data.
        """
        return encoder._encode_xarray(data=self.data, **kwargs)


class XArrayDatasetWrapper(XArrayDataArrayWrapper):
    """Wrapper around an xarray `DataSet`, offering polymorphism and convenience
    methods.
    """

    def to_numpy(self, flatten=False):
        """Return a numpy `ndarray` representation of the data.

        Returns
        -------
        numpy.ndarray
        """
        arr = self.data.to_array().to_numpy()
        if flatten:
            arr = arr.flatten()
        return arr

    # def component(self, component):
    #     """
    #     Get the data representing a specific vector component.

    #     Parameters
    #     ----------
    #     component : str
    #         The vector component to extract from the data. Accepts values of
    #         `u` or `v` (case-insensitive).

    #     Returns
    #     -------
    #     xarray.core.dataarray.DataArray
    #         An xarray `DataArray` containing the data representing the given
    #         component.
    #     """
    #     candidates = COMPONENTS.get(component, [])
    #     for variable in candidates:
    #         if variable in self.source.data_vars:
    #             break
    #     else:
    #         raise ValueError(f"No variable found with direction '{component}'")
    #     return self.source.data_vars[variable]


def wrapper(data, *args, fieldlist=True, **kwargs):
    from earthkit.data.utils import is_module_loaded

    if not is_module_loaded("xarray"):
        return None

    import xarray as xr

    ds = None
    if isinstance(data, xr.Dataset):
        ds = data
    elif isinstance(data, xr.DataArray):
        try:
            ds = data.to_dataset()
        except ValueError:
            return XArrayDataArrayWrapper(data, *args, **kwargs)

    if ds is not None:
        if not fieldlist:
            return XArrayDatasetWrapper(ds, *args, **kwargs)

        from earthkit.data.readers.netcdf.fieldlist import XArrayFieldList

        fs = XArrayFieldList(ds, **kwargs)
        if fs.has_fields():
            return fs
        else:
            return XArrayDatasetWrapper(ds, *args, **kwargs)

    return None



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


def test_odb_to_pandas():
    ds = from_source("file", earthkit_examples_file("test.odb"))
    df = ds.to_pandas()
    assert len(df) == 717
    cols = [
        "lat",
        "lon",
        "fg_dep",
        "an_dep",
    ]
    assert list(df.columns) == cols
    assert np.allclose(df["lat"][0:2], [38.808998, 73.6931])
    assert np.allclose(df["lon"][0:2], [4.2926, -3.4167])
    assert np.allclose(df["fg_dep"][0:2], [0.514543, -0.098977])


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import numpy as np
import pytest

from earthkit.data import config
from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file


@pytest.mark.long_test
@pytest.mark.cache
@pytest.mark.parametrize(
    "field_policy",
    [
        {"grib-field-policy": "persistent"},
        {"grib-field-policy": "temporary"},
    ],
)
@pytest.mark.parametrize(
    "handle_policy",
    [
        {"grib-handle-policy": "persistent"},
        {"grib-handle-policy": "temporary"},
        {"grib-handle-policy": "cache", "grib-handle-cache-size": 1},
        {"grib-handle-policy": "cache", "grib-handle-cache-size": 10},
        {"grib-handle-policy": "cache", "grib-handle-cache-size": 100},
        {"grib-handle-policy": "cache", "grib-handle-cache-size": 300},
    ],
)
@pytest.mark.parametrize(
    "_kwargs",
    [
        {},
        {"chunks": "auto"},
        {"chunks": {"valid_time": 1}},
        {"chunks": {"valid_time": 10}},
        {"chunks": {"valid_time": (100, 200, 432), "latitude": (1, 2), "longitude": (2, 1)}},
        {"chunks": -1},
    ],
)
def test_xr_engine_chunk_1(field_policy, handle_policy, _kwargs):
    with config.temporary(**field_policy, **handle_policy):
        ds_in = from_source(
            "url", earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_1_year.grib")
        )

        ds = ds_in.to_xarray(time_dim_mode="valid_time", **_kwargs)

        assert ds is not None

        r = ds["2t"].mean("valid_time").load()

        assert np.isclose(r.values.mean(), 287.2627299620878)


# This test is a copy of the previous one, but only using the default config
@pytest.mark.cache
@pytest.mark.parametrize(
    "_kwargs",
    [
        {},
        {"chunks": "auto"},
        {"chunks": {"valid_time": 1}},
        {"chunks": {"valid_time": 10}},
        {"chunks": {"valid_time": (100, 200, 432), "latitude": (1, 2), "longitude": (2, 1)}},
        {"chunks": -1},
    ],
)
def test_xr_engine_chunk_2(_kwargs):
    # the default settings
    field_policy = {"grib-field-policy": "persistent"}
    handle_policy = {"grib-handle-policy": "cache", "grib-handle-cache-size": 1}

    with config.temporary(**field_policy, **handle_policy):
        ds_in = from_source(
            "url", earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_1_year.grib")
        )

        ds = ds_in.to_xarray(time_dim_mode="valid_time", **_kwargs)

        assert ds is not None

        r = ds["2t"].mean("valid_time").load()

        assert np.isclose(r.values.mean(), 287.2627299620878)


@pytest.mark.cache
@pytest.mark.parametrize(
    "_kwargs",
    [
        {},
        {"chunks": "auto"},
        {"chunks": {"valid_time": 1}},
        {"chunks": {"valid_time": 10}},
        {"chunks": {"valid_time": (100, 200, 432), "latitude": (1, 2), "longitude": (2, 1)}},
        {"chunks": {"valid_time": 100, "latitude": 2, "longitude": 1}},
        {"chunks": -1},
    ],
)
def test_xr_engine_chunk_3(_kwargs):
    # in-memory fieldlist
    ds_in = from_source(
        "url",
        earthkit_remote_test_data_file("test-data", "xr_engine", "date", "t2_1_year.grib"),
        stream=True,
        read_all=True,
    )

    ds = ds_in.to_xarray(time_dim_mode="valid_time", **_kwargs)
    assert ds is not None

    r = ds["2t"].mean("valid_time").load()

    assert np.isclose(r.values.mean(), 287.2627299620878)



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

ds_xr = ds.to_xarray()

# option1: writing to GRIB file using the accessor
ds_xr.earthkit.to_target("file", "_res_xarray_to_grib_2.grib")

# option2: writing to GRIB file using the top level function
ekd.to_target("file", "_res_xarray_to_grib_2a.grib", data=ds_xr)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from abc import ABCMeta
from abc import abstractmethod
from functools import cached_property

LOG = logging.getLogger(__name__)


class RequestMapper(metaclass=ABCMeta):
    metadata_alias = None

    def __init__(self, request, **kwargs):
        self.request = request

    @cached_property
    def field_requests(self):
        return self._build()

    @abstractmethod
    def _build(self):
        pass

    def request_at(self, index):
        return self.field_requests[index]

    def __len__(self):
        return len(self.field_requests)



#!/usr/bin/env python3

# (C) Copyright 2024 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.readers.geotiff import GeoTIFFField
from earthkit.data.testing import NO_RIOXARRAY
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils.projections import TransverseMercator


@pytest.mark.skipif(NO_RIOXARRAY, reason="rioxarray not available")
@pytest.mark.with_proj
@pytest.mark.parametrize(
    "fname", [("dgm50hs_col_32_368_5616_nw.tif"), ("dgm50hs_col_32_368_5616_nw_bigtiff.tiff")]
)
def test_geotiff_reader_with_multiband(fname):
    s = from_source("file", earthkit_test_data_file(fname))
    assert len(s) == 3
    assert isinstance(s[0], GeoTIFFField)
    assert isinstance(s[1], GeoTIFFField)
    assert s[0].metadata("band") == 1
    assert s[1].metadata("band") == 2
    assert s[2].metadata("band") == 3
    assert isinstance(s.projection(), TransverseMercator)
    assert s[0].shape == (294, 315)


@pytest.mark.skipif(NO_RIOXARRAY, reason="rioxarray not available")
@pytest.mark.with_proj
def test_geotiff_bypassing_xr_engine():
    import xarray as xr

    # the earthkit xarray engine should not try to handle this file but leave it to the rioxarray backend
    da = xr.open_dataarray(earthkit_test_data_file("dgm50hs_col_32_368_5616_nw.tif"))
    assert da.name == "band_data"


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data import from_source
from earthkit.data.decorators import normalize
from earthkit.data.testing import earthkit_examples_file


def f(d):
    return d


@pytest.mark.xfail
def test_normalize_dates_from_source():
    dates_3 = normalize("d", "date")(f)
    dates_list_3 = normalize("d", "date", multiple=True)(f)

    source = from_source("file", earthkit_examples_file("test.grib"))
    assert dates_3(source[0]) == datetime.datetime(2020, 5, 13, 12, 0)
    assert dates_list_3(source[0]) == [datetime.datetime(2020, 5, 13, 12, 0)]

    source = from_source("file", earthkit_examples_file("test.nc"))

    #  For now
    with pytest.raises(NotImplementedError):
        assert dates_3(source[0]) == datetime.datetime(2020, 5, 13, 12, 0)
        assert dates_list_3(source[0]) == [datetime.datetime(2020, 5, 13, 12, 0)]


def test_normalize_dates_formatted_1():
    date_formatted = normalize("d", "date", format="%Y.%m.%d")(f)

    assert date_formatted("20200513") == "2020.05.13"


@pytest.mark.skip(reason="Not implemented yet.")
def test_normalize_enum_dates_formatted():
    date_formatted = normalize("d", values=["20010512", "20020512"], type="date", format="%Y.%m.%d")(f)

    assert date_formatted("20200513") == "2020.05.13"


def test_normalize_dates_formatted():
    date_formatted = normalize("d", "date-list(%Y.%m.%d)")(f)

    assert date_formatted(["20200513", "20200514"]) == ["2020.05.13", "2020.05.14"]
    assert date_formatted("20200513") == ["2020.05.13"]
    assert date_formatted([datetime.datetime(2020, 5, 13, 0, 0)]) == ["2020.05.13"]
    assert date_formatted([datetime.datetime(2020, 5, 13, 23, 59)]) == ["2020.05.13"]


def test_normalize_dates_multiple():
    date_1 = normalize("d", "date-list(%Y.%m.%d)")(f)
    date_2 = normalize("d", "date(%Y.%m.%d)", multiple=True)(f)
    date_3 = normalize("d", "date(%Y.%m.%d)", multiple=False)(f)
    date_4 = normalize("d", "date-list(%Y.%m.%d)", multiple=False)(f)

    assert date_1("20200511") == ["2020.05.11"]
    assert date_2("20200512") == ["2020.05.12"]
    assert date_3("20200513") == "2020.05.13"

    with pytest.raises(ValueError):
        date_4("20200514")


@pytest.mark.parametrize(
    "date,expected",
    [
        (
            "20200511/20200512",
            [
                "2020.05.11",
                "2020.05.12",
            ],
        ),
        (
            "20200511/20200514",
            [
                "2020.05.11",
                "2020.05.14",
            ],
        ),
        (
            "20200511/TO/20200512",
            [
                "2020.05.11",
                "2020.05.12",
            ],
        ),
        (
            "20200511/to/20200512",
            [
                "2020.05.11",
                "2020.05.12",
            ],
        ),
        (
            "20200511/TO/20200513/BY/2",
            [
                "2020.05.11",
                "2020.05.13",
            ],
        ),
        (
            "20200511/to/20200513/by/2",
            [
                "2020.05.11",
                "2020.05.13",
            ],
        ),
        (
            ["20200511", "20200513"],
            [
                "2020.05.11",
                "2020.05.13",
            ],
        ),
        (
            ["2020-05-11", "2020-05-13"],
            [
                "2020.05.11",
                "2020.05.13",
            ],
        ),
        (
            "20200511/to/20200513",
            [
                "2020.05.11",
                "2020.05.12",
                "2020.05.13",
            ],
        ),
        (
            ["20200511", "20200512", "20200513"],
            [
                "2020.05.11",
                "2020.05.12",
                "2020.05.13",
            ],
        ),
        (
            "20200511/to/20200515",
            [
                "2020.05.11",
                "2020.05.12",
                "2020.05.13",
                "2020.05.14",
                "2020.05.15",
            ],
        ),
        (
            ["20200511", "20200512", "20200513", "20200514", "20200515"],
            [
                "2020.05.11",
                "2020.05.12",
                "2020.05.13",
                "2020.05.14",
                "2020.05.15",
            ],
        ),
    ],
)
def test_normalize_date_list(date, expected):
    n = normalize("d", "date-list(%Y.%m.%d)")(f)
    assert n(date) == expected


@pytest.mark.xfail
def test_normalize_dates_formatted_from_pandas():
    import pandas as pd

    df1 = pd.DataFrame(
        [
            datetime.datetime(2005, 8, 27, 18, 0),
        ],
        columns=["date"],
    )
    df2 = pd.DataFrame(
        [
            datetime.datetime(2005, 8, 26, 18, 0),
            datetime.datetime(2005, 8, 27, 18, 0),
        ],
        columns=["date"],
    )

    @normalize("date", "date-list(%Y-%m-%d)")
    def f(date):
        return date

    print(f(df1))
    print(f(df2))

    @normalize("date", "date(%Y-%m-%d)")
    def f(date):
        return date

    print(f(df1))
    with pytest.raises(AssertionError):
        print(f(df2))


@pytest.mark.skip("Not implemented (yet?).")
def test_normalize_dates_formatted_from_object():
    date_formatted = normalize("d", "date", format="%Y.%m.%d")(f)

    class CustomDateObject:
        def __init__(self, dates):
            self.dates = dates

        def to_datetime_list(self):
            return self.dates

    obj = CustomDateObject(
        [
            datetime.datetime(2005, 8, 26, 18, 0),
            datetime.datetime(2005, 8, 26, 18, 0),
        ]
    )

    assert date_formatted(obj) == "2020.05.13"


def test_normalize_date_none_1():
    @normalize(
        "name",
        "date(%Y%m%d)",
    )
    def date_default_none(name=None):
        return name

    assert date_default_none("2012-12-02") == "20121202"
    assert date_default_none() is None


def test_normalize_date_list_none_1():
    @normalize(
        "name",
        "date-list(%Y%m%d)",
    )
    def date_default_none(name=None):
        return name

    assert date_default_none("2012-12-02") == ["20121202"]
    assert date_default_none() is None


def test_normalize_date_default_1():
    @normalize(
        "name",
        "date",
    )
    def date_default_1(name="wrong-default"):
        return name

    date_default_1("2012-12-02")
    with pytest.raises(ValueError, match=".*wrong-default.*"):
        date_default_1()


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)
